import re
import tifffile
from bigfish.stack import read_image
from czifile import imread
from czifile import CziFile
from ..utils import check_parameter
from typing import Optional, Tuple

class FormatError(Exception):
    pass


def open_image(full_path:str) :
    if full_path.endswith('.czi') : im = imread(full_path)
    else : im = read_image(full_path)

    reshape = []
    for axis in im.shape :
        if axis != 1 : reshape.append(axis)
    im = im.reshape(reshape)

    return im


def check_format(image, is_3D, is_multichannel) :
    shape = list(image.shape)
    dim = image.ndim - (shape[image.ndim - 1] == 1)
    if not dim == (2 + is_3D  + is_multichannel) :
        raise FormatError("Inconsistency in image format and parameters.")



def get_filename(full_path: str) :
    check_parameter(full_path=str)

    pattern = r'.*\/(.+)\..*$'
    if not full_path.startswith('/') : full_path = '/' + full_path
    re_match = re.findall(pattern, full_path)
    if len(re_match) == 0 : raise ValueError("Could not read filename from image full path.")
    if len(re_match) == 1 : return re_match[0]
    else : raise AssertionError("Several filenames read from path")

def get_voxel_size(filepath: str) -> Optional[Tuple[Optional[float], Optional[float], Optional[float]]]:
    """
    Returns voxel size in nanometers (nm) as a tuple (X, Y, Z).
    Any of the dimensions may be None if not available.
    /WARINING\\ : the unit might not be nm
    """
    try:
        if filepath.endswith('.czi'):
            with CziFile(filepath) as czi:
                metadata = czi.metadata()  # returns XML metadata
                # try to parse voxel sizes from XML
                import xml.etree.ElementTree as ET
                root = ET.fromstring(metadata)
                scaling_distance = root.findall('.//Scaling//Items//Distance//Value')
                if len(scaling_distance) in [2,3] :
                    for scale in scaling_distance :
                        res = [float(scale.text) * 1e9 for scale in scaling_distance] #m to nm
                    res.reverse()
                    return tuple(res)
                else :
                    raise Exception("Couln't find voxel size on xml metadata")

        elif filepath.endswith(('.tif', '.tiff')):
            with tifffile.TiffFile(filepath) as tif:
                ij_meta = tif.imagej_metadata
                page = tif.pages[0]  # first image page
                # X/Y resolution as (numerator, denominator)
                xres = page.tags['XResolution'].value
                yres = page.tags['YResolution'].value
                # ResolutionUnit: must be 'nm' for this calculation
                res_unit = ij_meta.get("unit")

                if res_unit and str(res_unit) != 'nm':
                    xy_size = 1 / (xres[0] / xres[1]) * 1e3 #um to nm
                elif res_unit and str(res_unit) != 'NONE':
                    xy_size = 1 / (xres[0] / xres[1]) #um to nm
                else:
                    xy_size = None

                # Z spacing from ImageJ metadata
                if res_unit and str(res_unit) != 'nm':
                    z_size = ij_meta.get('spacing', None) * 1e3 
                else :
                    z_size = ij_meta.get('spacing', None)

                return (z_size,xy_size, xy_size )
    except Exception as e:
        print(f"Failed to read voxel size from {filepath}: {e}")
        return None


import FreeSimpleGUI as sg
import os
import numpy as np
import cellpose.models as models
from typing import Optional, Union

from cellpose.core import use_gpu
from .tooltips import FLOW_THRESHOLD_TOOLTIP,CELLPROB_TOOLTIP, MIN_SIZE_TOOLTIP
from ..hints import pipeline_parameters
from ..utils import check_parameter
from ..interface import SettingsDict, get_default_settings, get_settings


settings = get_settings()

def add_header(header_text) :
    """Returns [elmnt] not layout"""
    header = [sg.Text('\n{0}'.format(header_text), size= (len(header_text),3), font= 'bold 15')]
    return header


def pad_right(string, length, pad_char) :
    if len(string) >= length : return string
    else : return string + pad_char* (length - len(string))
    

def parameters_layout(
        parameters:'list[str]' = [], 
        unit=None, 
        header= None, 
        default_values=None, 
        keys = None,
        tooltips = None,
        size=5, 
        opt:list=None
        ) :

    if len(parameters) == 0 : return []
    check_parameter(parameters= list, header = (str, type(None)))
    for key in parameters : check_parameter(key = str)
    max_length = len(max(parameters, key=len))

    if keys is None : keys = [None] * len(parameters)
    else :
        if len(keys) != len(parameters) : raise ValueError("keys length must be equal with parameters length or set to None.")
    if tooltips is None : tooltips = [None] * len(parameters)
    else :
        if len(tooltips) != len(parameters) : raise ValueError("tooltips length must be equal with parameters length or set to None.")

    if type(opt) == type(None) :
        opt = [False] * len(parameters)
    else :
        if len(opt) != len(parameters) : raise ValueError("Parameters and opt must be of same length.")

    if isinstance(default_values, (list, tuple)) :
        if len(default_values) != len(parameters) : raise ValueError("if default values specified it must be of equal length as parameters.")
        layout= [
            [
                sg.Text("{0}".format(pad_right(parameter, max_length, ' ')), text_color= 'green' if option else None), 
                sg.InputText(size= size, key= parameter if key is None else key, default_text= value, tooltip=tooltip)
            
            ] for parameter,value, option, key, tooltip in zip(parameters,default_values, opt, keys, tooltips)
        ]
    else :
        layout= [
            [sg.Text("{0}".format(pad_right(parameter, max_length, ' ')), text_color= 'green' if option else None), 
             sg.InputText(size= size, key= parameter)
             
             ] for parameter, option in zip(parameters, opt)
        ]
    
    if isinstance(unit, str):
        for line_id, line in enumerate(layout) :
            layout[line_id] += [sg.Text('{0}'.format(unit))]
    elif isinstance(unit, list) :
        if len(unit) != len(parameters) : raise ValueError(f"unit list and parameters must have same length : {len(unit)} , {len(parameters)}")

        for line_id, unit_txt in zip(range(len(parameters)), unit) :
            layout[line_id] += [sg.Text('{0}'.format(unit_txt))]
    
    if isinstance(header, str) :
        layout = [add_header(header)] + layout
    return layout

def tuple_layout(opt=None, default_dict={}, unit:dict={}, names : dict = {}, **tuples) :
    """
    tuples example : voxel_size = ['z','y','x']; will ask a tuple with 3 element default to 'z', 'y' 'x'.
    """
    if type(tuples) == type(None) : return []
    if len(tuples.keys()) == 0 : return []

    if type(opt) != type(None) :
        if not isinstance(opt, dict) : raise TypeError("opt parameter should be either None or dict type.")
        if not opt.keys() == tuples.keys() : raise ValueError("If opt is passed it is expected to have same keys as tuples dict.")
    else : 
        opt = tuples.copy()
        for key in opt.keys() :
            opt[key] = False

    for tup in tuples : 
        if not isinstance(tuples[tup], (list,tuple)) : raise TypeError()

    max_size = len(max(tuples.keys(), key=len))
    
    layout = [
        [sg.Text(pad_right(names[tup] if tup in names.keys() else tup, max_size, ' '), text_color= 'green' if opt[option] else None)] 
        + [sg.InputText(default_text=default_dict.setdefault('{0}_{1}'.format(tup,elmnt), elmnt),key= '{0}_{1}'.format(tup, elmnt), size= 5) for elmnt in tuples[tup]]
        + [sg.Text(unit.setdefault(tup,''))]
        for tup,option, in zip(tuples,opt)
    ]

    return layout

def path_layout(keys= [],look_for_dir = False, header=None, preset=settings.working_directory) :
    """
    If not look for dir then looks for file.
    """
    if len(keys) == 0 : return []
    check_parameter(keys= list, header = (str, type(None)))
    for key in keys : check_parameter(key = str)
    
    if look_for_dir : Browse = sg.FolderBrowse
    else : Browse = sg.FileBrowse

    max_length = len(max(keys, key=len))

    layout = []
    for name in keys :
        layout += [
            [sg.Text(pad_right(name, max_length, ' '))], 
            [sg.InputText(key= name, expand_x=True, default_text=preset), Browse(key= name + "_browse", initial_folder= preset), sg.Text('')],
            ]
    if isinstance(header, str) :
        layout = [add_header(header)] + layout
    return layout

def bool_layout(parameters= [], header=None, preset : Optional[Union['list[bool]',bool,None]]=None, keys=None) :
    if len(parameters) == 0 : return []
    check_parameter(parameters= list, header= (str, type(None)), preset=(type(None), list, tuple, bool))
    for key in parameters : check_parameter(key = str)
    if type(preset) == type(None) :
        preset = [False] * len(parameters)
    elif type(preset) == bool :
        preset = [preset] * len(parameters)
    else : 
        for key in preset : check_parameter(key = bool)

    max_length = len(max(parameters, key=len))

    if type(keys) == type(None) :
        keys = parameters
    elif isinstance(keys,(list,tuple)) :
        if len(keys) != len(parameters) : raise ValueError("keys arguement must be of same length than parameters argument")
    elif len(parameters) == 1 and isinstance(keys, str) :
        keys = [keys]
    else :
        raise ValueError('Incorrect keys parameters. Expected list of same length than parameters or None.')

    layout = [
        [sg.Checkbox(pad_right(name, max_length, ' '), key=key, default=box_preset)] for name, box_preset, key in zip(parameters,preset,keys)
    ]
    if isinstance(header, str) :
        layout = [add_header(header)] + layout
    return layout

def combo_elmt(values, key, header=None, read_only=True, default_value=None) :
    """
    drop-down list
    """
    if len(values) == 0 : return []
    check_parameter(values= list, header= (str, type(None)))
    if type(default_value) == type(None) :
        default_value = values[0]
    elif default_value not in values :
        default_value = values[0]
    layout = [
        sg.Combo(values, default_value=default_value, readonly=read_only, key=key)
    ]
    if isinstance(header, str) :
        layout = add_header(header) + layout
    return layout

def radio_layout(values, header=None, key=None) :
    """
    Single choice buttons.
    """
    if len(values) == 0 : return []
    check_parameter(values= list, header= (str, type(None)))
    layout = [
        [sg.Radio(value, group_id= 0, key=key) for value in values]
    ]
    if isinstance(header, str) :
        layout = [add_header(header)] + layout
    return layout

def _segmentation_layout(
        is_multichannel : bool,
        is_3D_stack : bool,
        cytoplasm_model,
        nucleus_model,
        cytoplasm_channel,
        nucleus_channel,
        other_nucleus_image,
        cytoplasm_diameter,
        nucleus_diameter,
        show_segmentation,
        save_segmentation_visuals, 
        segment_only_nuclei,
        saving_path,
        filename,
        cytoplasm_segmentation_3D ,
        nucleus_segmentation_3D ,
        cellprob_threshold ,
        flow_threshold ,
        anisotropy,
        cytoplasm_min_size : int,
        nucleus_min_size : int,
        reordered_shape : tuple,
        **kwargs
        ) :
    
    USE_GPU = use_gpu()
    event_dict = dict()
    
    #Header : GPU availabality
    layout = [[sg.Text("Cell Segmentation", font="bold 20", pad=(0,20))]]
    layout += [[sg.Text("GPU is currently "), sg.Text('ON', text_color= 'green') if USE_GPU else sg.Text('OFF', text_color= 'red')]]
    layout += bool_layout(['Interactive segmentation'],keys=['show_segmentation'], preset= show_segmentation)
    
    segment_only_nuclei_checkbox = sg.Checkbox("Segment only nuclei", default=segment_only_nuclei, key= "segment_only_nuclei", enable_events=True) 
    layout += [[segment_only_nuclei_checkbox]]
    event_dict['segment_only_nuclei'] = segment_only_nuclei_checkbox
    
    #Nucleus parameters
    nucleus_key = "nucleus"
    layout += [[sg.Text("Nucleus parameters", font="bold 15", pad=(0,10))]]
    nucleus_parameters_col, nucleus_event_dict = _segmentate_object_layout(
        reordered_shape = reordered_shape,
        anisotropy=anisotropy,
        cellprob_threshold=cellprob_threshold,
        channel=nucleus_channel,
        diameter=nucleus_diameter,
        flow_threshold=flow_threshold,
        is_3D_stack=is_3D_stack,
        model=nucleus_model,
        is_multichannel=is_multichannel,
        object_key=nucleus_key,
        segmentation_3D=nucleus_segmentation_3D,
        min_size=nucleus_min_size
    )
    layout += path_layout(keys=["other_nucleus_image"], look_for_dir=False, preset=other_nucleus_image)
    layout += [[nucleus_parameters_col]]
    event_dict.update(nucleus_event_dict)
    event_dict[nucleus_key + "_column"] = nucleus_parameters_col


    #Cytoplasm parameters
    layout += [[sg.Text("Cytoplasm parameters", font="bold 15", pad=(0,10))]]
    cytoplasm_key = "cytoplasm"
    cytoplasm_parameters_col, cytoplasm_event_dict = _segmentate_object_layout(
        reordered_shape = reordered_shape,
        anisotropy=anisotropy,
        cellprob_threshold=cellprob_threshold,
        channel=cytoplasm_channel,
        diameter=cytoplasm_diameter,
        flow_threshold=flow_threshold,
        is_3D_stack=is_3D_stack,
        model=cytoplasm_model,
        is_multichannel=is_multichannel,
        object_key=cytoplasm_key,
        segmentation_3D=cytoplasm_segmentation_3D,
        min_size=cytoplasm_min_size
    )
    layout += [[cytoplasm_parameters_col]]
    event_dict.update(cytoplasm_event_dict)
    event_dict[cytoplasm_key + "_column"] = cytoplasm_parameters_col

    #Control plots
    layout += [[sg.Text("Control plots", font="bold 15", pad=(0,10))]]
    layout += bool_layout(['Save png control'], preset= save_segmentation_visuals, keys=['save_segmentation_visuals'])
    layout += path_layout(keys=['seg_control_saving_path'], look_for_dir=True, preset=saving_path)
    layout += parameters_layout(['Filename'], default_values=[filename], size= 25, keys=['filename'])

    return layout, event_dict

def _segmentate_object_layout(
    reordered_shape : tuple | None,
    object_key : str,     
    is_multichannel : bool,
    is_3D_stack : bool,
    model : str,
    channel : int,
    diameter : int,
    segmentation_3D : bool ,
    cellprob_threshold : float ,
    flow_threshold : float,
    anisotropy : float,
    min_size : int,
    **kwargs
    ) -> sg.Column :

    models_list = models.get_user_models() + models.MODEL_NAMES
    if len(models_list) == 0 : models_list = ['no model found']

    key_2D = object_key + "_radio_2D"
    options_2D = list()
    key_3D = object_key + "_radio_3D"
    options_3D = list()
    layout = []

    radio_2D_seg = sg.Radio("2D segmentation", group_id=object_key+"_seg_dim", default=not segmentation_3D, visible = True, enable_events=True, key=key_2D)
    radio_max_proj = sg.Radio("max proj", group_id=object_key+"_2D_proj", default=False, disabled= segmentation_3D, key= object_key+"_max_proj")
    radio_mean_proj = sg.Radio("mean proj", group_id=object_key+"_2D_proj", default=True, disabled= segmentation_3D, key = object_key + "_mean_proj")
    radio_slice_proj = sg.Radio("select slice", group_id=object_key+"_2D_proj", default=False, disabled= segmentation_3D, key=object_key + "_select_slice")
    radio_3D_seg = sg.Radio("3D segmentation", group_id=object_key+"_seg_dim", default=segmentation_3D, visible = True, enable_events=True, key= key_3D)
    anisotropy = sg.Input(default_text = 1, key = object_key + "_anisotropy",size=5, disabled=not segmentation_3D)

    if is_3D_stack :        
        slice_number = reordered_shape[0 + is_multichannel] if not reordered_shape is None else 999
        int_slice_proj = sg.Spin(list(range(slice_number)), size= (5,1), disabled= segmentation_3D, key=object_key+"_selected_slice")

        options_2D += [
            radio_max_proj, 
            radio_mean_proj, 
            radio_slice_proj, 
            int_slice_proj
            ]

        layout += [
                [radio_2D_seg],
                [sg.Column([[radio_max_proj, radio_mean_proj, radio_slice_proj, int_slice_proj]], pad= (15,0,0,5))]
                ]


        layout += [
            [radio_3D_seg],
            [sg.Column([[sg.Text("anisotropy"), anisotropy]], pad=(15,0,0,5))]
            ]  
        
        options_3D += [
            anisotropy
        ]

    if is_multichannel : 
        channel_elmt = sg.Input(channel, key=object_key + "_channel", size=5)
        layout += [[sg.Text(f'{object_key.capitalize()} channel'), channel_elmt]]

    layout += [[sg.Text("Cellpose model : ")] + combo_elmt(models_list, key=object_key +'_model_name', default_value= model)]
    layout += parameters_layout([f'{object_key.capitalize()} diameter'], unit= "px", default_values= [diameter], keys=[object_key+'_diameter'])
    layout += parameters_layout(
        ["Flow threshold", "Cellprob threshold", "Min. size"],
        unit=["","","px"], 
        default_values=[flow_threshold, cellprob_threshold, min_size], 
        keys=[object_key + "_flow_threshold",object_key + "_cellprob_threshold", object_key + "_min_size"],
        tooltips= [FLOW_THRESHOLD_TOOLTIP, CELLPROB_TOOLTIP, MIN_SIZE_TOOLTIP]
        )

    object_col = sg.Column(layout)

    #Reference dict
    event_dict = {
        key_2D : options_2D,
        key_3D : options_3D,
        object_key + "_radio_2D_seg" : radio_2D_seg,
        object_key + "_radio_3D_seg" : radio_3D_seg,
        object_key + "_channel" : channel_elmt, # For batch mode layout update
    }

    return object_col, event_dict
    

def _input_parameters_layout(
        ask_for_segmentation : bool,
        is_3D_stack_preset : bool,
        time_stack_preset : bool,
        multichannel_preset : bool,
        do_dense_regions_deconvolution_preset : bool,
        do_clustering_preset : bool,
        do_segmentation_preset : bool,
        do_Napari_correction : bool
    ) :
    layout_image_path = path_layout(['image_path'], header= "Image")
    layout_image_path += bool_layout(['3D stack', 'Multichannel stack'], keys=['is_3D_stack', 'is_multichannel'], preset= [is_3D_stack_preset, multichannel_preset])
    
    layout_image_path += bool_layout(
        ['Dense regions deconvolution', 'Compute clusters', 'Cell segmentation', 'Open Napari corrector'],
        keys= ['do_dense_regions_deconvolution', 'do_cluster_computation', 'do_segmentation','show_napari_corrector' ], 
        preset= [do_dense_regions_deconvolution_preset, do_clustering_preset, do_segmentation_preset, do_Napari_correction], 
        header= "Pipeline settings")


    return layout_image_path

def _detection_layout(
        is_3D_stack,
        is_multichannel,
        do_dense_region_deconvolution,
        do_clustering,
        do_segmentation,
        segmentation_done=False,
        default_dict : pipeline_parameters={},
    ) :
    if is_3D_stack : dim = 3
    else : dim = 2
    default = get_settings()

    #Detection
    detection_parameters = ['threshold', 'threshold penalty']
    default_detection = [default_dict.setdefault('threshold',default.threshold), default_dict.setdefault('threshold penalty', default.threshold_penalty)]
    opt= [True, True]
    parameters_keys = ['threshold', 'threshold_penalty']
    if is_multichannel : 
        detection_parameters += ['channel to compute']
        opt += [False]
        parameters_keys += ['channel_to_compute']
        default_detection += [default_dict.setdefault('channel_to_compute', default.detection_channel)]
    
    layout = [[sg.Text("Green parameters", text_color= 'green'), sg.Text(" are optional parameters.")]]
    layout += parameters_layout(detection_parameters, header= 'Detection', opt=opt, default_values=default_detection, keys= parameters_keys)
    
    if dim == 2 : tuple_shape = ('y','x')
    else : tuple_shape = ('z','y','x')
    opt = {'voxel_size' : False, 'spot_size' : False, 'log_kernel_size' : True, 'minimum_distance' : True}
    unit = {'voxel_size' : 'nm', 'minimum_distance' : 'nm', 'spot_size' : 'radius(nm)', 'log_kernel_size' : 'px'}
    names = {'voxel_size' : 'Voxel size', 'spot_size' : 'Spot size', 'log_kernel_size' : 'LoG kernel size', 'minimum_distance' : 'Minimum distance between spots'}

    layout += tuple_layout(opt=opt, unit=unit, default_dict=default_dict, names=names, voxel_size= tuple_shape, spot_size= tuple_shape, log_kernel_size= tuple_shape, minimum_distance= tuple_shape)
    
    if (do_segmentation and is_multichannel) or (is_multichannel and segmentation_done):
        layout += [[sg.Text("nucleus channel signal "), sg.InputText(default_text=default_dict.setdefault('nucleus_channel',default.nucleus_channel), key= "nucleus channel signal", size= 5, tooltip= "Channel from which signal will be measured for nucleus features, \nallowing you to measure signal from a different channel than the one used for segmentation.")]]
    layout += bool_layout(['Interactive threshold selector'],keys = ['show_interactive_threshold_selector'], preset=[default.interactive_threshold_selector])
    
    #Deconvolution
    if do_dense_region_deconvolution :
        default_dense_regions_deconvolution = [default_dict.setdefault('alpha',default.alpha), default_dict.setdefault('beta',default.beta)]
        layout += parameters_layout(['alpha', 'beta',], default_values= default_dense_regions_deconvolution, header= 'Dense regions deconvolution')
        layout += parameters_layout(['gamma'], unit= 'px', default_values= [default_dict.setdefault('gamma',default.gamma)])
        layout += tuple_layout(opt= {"deconvolution_kernel" : True}, unit= {"deconvolution_kernel" : 'px'}, default_dict=default_dict, deconvolution_kernel = tuple_shape)
    
    #Clustering
    if do_clustering :
        layout += parameters_layout(['Cluster radius'],keys=['cluster_size'], unit="radius(nm)", default_values=[default_dict.setdefault('cluster_size',default.cluster_size)])
        layout += parameters_layout(['Min nb spots per cluster'],keys=['min_number_of_spots'], default_values=[default_dict.setdefault('min_number_of_spots', default.min_spot)])

    layout += path_layout(
        keys=['spots_extraction_folder'],
        look_for_dir=True,
        header= "Individual spot extraction",
        preset= default_dict.setdefault('spots_extraction_folder', default.spot_extraction_folder)
    )
    default_filename = default_dict.setdefault("filename","") + "_spot_extraction"
    layout += parameters_layout(
        parameters=['spots_filename'],
        default_values=[default_filename],
        size= 13
    )
    layout += bool_layout(
        ['.csv','.excel',],
        keys= ['do_spots_csv', 'do_spots_excel'],
        preset= [default_dict.setdefault('do_spots_csv',default.do_csv), default_dict.setdefault('do_spots_excel',default.do_excel),]
    )

    return layout

def colocalization_layout(spot_list : list, **default_values) :
    default = get_settings()

    element_dict = {}
    can_use_memory = len(spot_list) != 0

    for spot_id in [1,2] :
        element_dict[f"radio_spots{spot_id}_memory"] = sg.Radio("From memory : ", group_id = spot_id, enable_events = True, key = f"radio_spots{spot_id}_memory", default= can_use_memory or default_values.get(f"radio_spots{spot_id}_memory"), disabled= not can_use_memory)
        element_dict[f"radio_spots{spot_id}_load"] = sg.Radio("Load spot detection : ", group_id = spot_id, enable_events = True, key = f"radio_spots{spot_id}_load", default= (not can_use_memory) or (default_values.get(f"radio_spots{spot_id}_load")))
        
        element_dict[f"spots{spot_id}_browse"] = [
            sg.Input(size=20, key= f"spots{spot_id}_browse", default_text=default_values.get(f"spots{spot_id}_browse"), disabled=can_use_memory and not default_values.get(f"radio_spots{spot_id}_load")),
            sg.FileBrowse(key=f"spots{spot_id}_browsebutton", initial_folder=default_values.setdefault(f"spots{spot_id}_browsebutton", default_values["working_directory"]), disabled=can_use_memory and not default_values.get(f"radio_spots{spot_id}_load")),
        ]
        element_dict[f"spots{spot_id}_voxel_size"] = [
            sg.Text("voxel size"),
            sg.Input(size= 5, key= f"z_voxel_size_spot{spot_id}", default_text= default_values.setdefault(f"z_voxel_size_spot{spot_id}", "z"), disabled=can_use_memory and not default_values.get(f"radio_spots{spot_id}_load")),
            sg.Input(size= 5, key= f"y_voxel_size_spot{spot_id}", default_text= default_values.setdefault(f"y_voxel_size_spot{spot_id}", "y"), disabled=can_use_memory and not default_values.get(f"radio_spots{spot_id}_load")),
            sg.Input(size= 5, key= f"x_voxel_size_spot{spot_id}", default_text= default_values.setdefault(f"x_voxel_size_spot{spot_id}", "x"), disabled=can_use_memory and not default_values.get(f"radio_spots{spot_id}_load")),
        ]
        
        #Ref for updating
        element_dict[f"options_spots{spot_id}_memory"] = sg.Col(
            [[sg.DropDown(values=[""] + spot_list, key=f"spots{spot_id}_dropdown", size= 10, disabled= not can_use_memory or default_values.get(f"radio_spots{spot_id}_load"), default_value= default_values.setdefault(f"spots{spot_id}_dropdown", ""))]]
            )
        element_dict[f"options_spots{spot_id}_load"] = sg.Col(
            [element_dict[f"spots{spot_id}_browse"], element_dict[f"spots{spot_id}_voxel_size"]]
            )

    col1 = sg.Col([
        [sg.Text("Spots 1", size = 10)],
        [element_dict["radio_spots1_memory"]],
        [element_dict["options_spots1_memory"]],
        [element_dict["radio_spots1_load"]],
        [element_dict["options_spots1_load"]],
    ])

    col2 = sg.Col([
        [sg.Text("Spots 2", size = 10)],
        [element_dict["radio_spots2_memory"]],
        [element_dict["options_spots2_memory"]],
        [element_dict["radio_spots2_load"]],
        [element_dict["options_spots2_load"]],
    ])

    layout = [
        [sg.Push(), sg.Text("Co-localization", size=50, font="bold"), sg.Push()],
        [sg.VPush()],
        [col1,col2]
    ]
    layout += parameters_layout(['colocalisation distance'], unit= 'nm', default_values= [default_values.setdefault('colocalisation_distance', default_values["coloc_range"])])
    layout += [[sg.Button("Ok", bind_return_key=True),sg.Button("Cancel"),sg.Push(),],
        [sg.VPush()]
    ]

    return layout, element_dict

def settings_layout(default_values : SettingsDict = get_default_settings()) :

    if not isinstance(default_values, SettingsDict) : raise TypeError(f"Incorect type for default_values : {type(default_values)}; expected SettingsDict")
    models_list = models.get_user_models() + models.MODEL_NAMES

    layout = [[sg.Text("Default values", font="ArialBold 20")]]
    layout += [[sg.Text("Default working directory"), sg.Input(default_text=default_values.working_directory, key= "working_directory"), sg.FolderBrowse()]]
    
    image_layout = [[sg.Text("Image", font="ArialBold 15")]]
    image_layout += bool_layout(['Multichannel stack', '3D stack'],preset= [default_values.multichannel_stack, default_values.stack_3D], keys=["multichannel_stack", "stack_3D"])
    image_layout += parameters_layout(['Detection channel', 'Nucleus channel'],default_values=[default_values.detection_channel, default_values.nucleus_channel], keys=['detection_channel', 'nucleus_channel'])

    segmentation_layout = [[sg.Text("Segmentation", font="ArialBold 15")],
                            [sg.Text("Cytoplasm model : "), sg.DropDown(models_list, default_value=default_values.cytoplasm_model, key= "cytoplasm_model")],
                            [sg.Radio(default= default_values.cytoplasm_mean_proj, group_id=0, key="cytoplasm_mean_proj", text= "mean proj"), sg.Radio(default=default_values.cytoplasm_max_proj, group_id=0, key="cytoplasm_max_proj", text= "max proj"), sg.Radio(default= default_values.cytoplasm_select_slice, group_id=0, key="cytoplasm_select_slice", text= "single slice"), sg.Input(default_values.cytoplasm_selected_slice, size=5, key= "cytoplasm_selected_slice")],
                            [sg.Text("Nucleus model : "), sg.DropDown(models_list, default_value=default_values.nucleus_model, key= "nucleus_model")],
                            [sg.Radio(default= default_values.nucleus_mean_proj, group_id=1, key="nucleus_mean_proj", text= "mean proj"), sg.Radio(default= default_values.nucleus_max_proj, group_id=1, key="nucleus_max_proj", text= "max proj"), sg.Radio(default= default_values.nucleus_select_slice, group_id=1, key="nucleus_select_slice", text= "single slice"), sg.Input(default_values.nucleus_selected_slice, size=5, key= "nucleus_selected_slice")],
                            ]
    segmentation_layout += parameters_layout(
        ["Flow threshold", "Cellprob threshold", "Cytoplasm diameter", "Cytoplasm min size", "Nucleus diameter", "Nucleus min size", "Anisotropy"],
        default_values= [default_values.flow_threshold, default_values.cellprob_threshold, default_values.cytoplasm_diameter, default_values.cytoplasm_min_size, default_values.nucleus_diameter, default_values.nucleus_min_size, default_values.anisotropy], 
        keys=["flow_threshold", "cellprob_threshold", "cytoplasm_diameter", "cytoplasm_min_size", "nucleus_diameter", "nucleus_min_size", "anisotropy"]) 
    segmentation_layout += bool_layout(
        ["show segmentation", "segment only nuclei", "do 3D segmentation", "save segmentation visual"], 
        preset=[default_values.show_segmentation, default_values.segment_only_nuclei, default_values.do_3D_segmentation, default_values.save_segmentation_visuals],
        keys= ["show_segmentation", "segment_only_nuclei", "do_3D_segmentation","save_segmentation_visuals"])

    detection_layout = [[sg.Text("Detection", font="ArialBold 15")]]
    detection_layout += parameters_layout(
        ["Threshold", "Threshold penalty"],
         default_values=[default_values.threshold, default_values.threshold_penalty],
         keys=["threshold", "threshold_penalty"])
    detection_layout += bool_layout(
        ["Dense regions deconvolution", "Cluster computation", "show napari corrector", "Autofloresnce background removal", "interactive threshold selector"],
        preset=[default_values.do_dense_regions_deconvolution, default_values.do_cluster, default_values.show_napari_corrector, default_values.do_background_removal, default_values.interactive_threshold_selector], 
        keys=["do_dense_regions_deconvolution", "do_cluster", "show_napari_corrector","do_background_removal", "interactive_threshold_selector"])

    deconvolution_layout = [[sg.Text("Dense regions deconvolution", font="ArialBold 15")]]
    deconvolution_layout += parameters_layout(
        ["alpha", "beta", "gamma"],
        default_values=[default_values.alpha, default_values.beta, default_values.gamma]
        )

    clustering_layout = [[sg.Text("Cluster computation", font="ArialBold 15")]]
    clustering_layout += parameters_layout(
        ["Cluster size", "Min spot number"],
        default_values=[default_values.cluster_size, default_values.min_spot], 
        keys=["cluster_size", "min_spot"])

    coloc_layout = [[sg.Text("Co-localization computation", font="ArialBold 15")]]
    coloc_layout += parameters_layout(['Co-localization range'],
    default_values=[default_values.coloc_range],
     keys=['coloc_range'])
    coloc_layout += tuple_layout(default_dict= {
        'voxel_size_z' : default_values.voxel_size[0],
        'voxel_size_y' : default_values.voxel_size[1],
        'voxel_size_x' : default_values.voxel_size[2],
        }, voxel_size=('z','y','x'))

    spot_extraction_layout = [[sg.Text("Spots extraction", font="ArialBold 15")]]
    spot_extraction_layout += bool_layout(
        ["do csv", "do excel"],
        preset=[default_values.do_csv, default_values.do_excel],
        keys=["do_csv", "do_excel"]
        )
    spot_extraction_layout += [[sg.Text("spot extraction folder"), sg.Input(default_text=default_values.spot_extraction_folder, key="spot_extraction_folder", size=7) ,sg.FolderBrowse()]]

    background_removing_layout = [[sg.Text("Background removing (batch)", font="ArialBold 15")]]
    background_removing_layout += [
        [sg.Checkbox("Remove background :", key= "do_background_removal", default= settings.do_background_removal)],
        [sg.Text("Background channel : "), sg.Input(settings.background_channel, key = "background_channel", size= 5)]
        ]

    layout += [[sg.Col(image_layout, vertical_alignment='top', expand_x = True), sg.Col(background_removing_layout, vertical_alignment='top', expand_x = True)]]
    layout += [[sg.Col(segmentation_layout, vertical_alignment='top', expand_x = True), sg.Col(detection_layout, vertical_alignment='top', expand_x = True)]]
    layout += [[sg.Col(deconvolution_layout, vertical_alignment='top', expand_x = True), sg.Col(clustering_layout, vertical_alignment='top', expand_x = True)]]
    layout += [[sg.Col(spot_extraction_layout, vertical_alignment='top', expand_x = True), sg.Col(coloc_layout, vertical_alignment='top', expand_x = True)]]

    return layout

def _ask_channel_map_layout(
        shape,
        is_3D_stack,
        is_multichannel,
        is_time_stack,
        preset_map={},
    ) :
    
    x = preset_map.setdefault('x',0)
    y = preset_map.setdefault('y',0)
    z = preset_map.setdefault('z',0)
    c = preset_map.setdefault('c',0)
    t = preset_map.setdefault('t',0)

    layout = [
            [sg.Text("Dimensions mapping", font= "bold 15"), sg.Text("Image shape : {0}".format(shape))]
        ]
    layout += parameters_layout(['x','y'], default_values=[x,y])
    if is_3D_stack : layout += parameters_layout(['z'], default_values=[z])
    if is_multichannel : layout += parameters_layout(['c'], default_values=[c])
    if is_time_stack : layout += parameters_layout(['t'], default_values=[t])

    return layout


""""
Constant submodule to have a common reference for parameters default values
"""
import os

WORKING_DIRECTORY = None

#Image
IS_MULTICHANNEL = False
IS_3D_STACK = False
CHANNEL = 0
NUC_CHANNEL = 1

#Segmentation
FLOW_THRESHOLD = 0.4
CELLPROB_THRESHOD = 0.
CYTO_MODEL = "cpsam"
NUC_MODEL = "cpsam"
CYTOPLASM_DIAMETER = 90
NUC_DIAMETER = 60
ANISOTROPY = 1. 
SHOW_SEGMENTATION = True
SEGMENT_ONLY_NUCLEI = False
DO_3D_SEMGENTATION = False
VISUAL_PATH = os.getcwd()
SAVE_SEGMENTATION_VISUAL = False
NUCLEUS_MIN_SIZE = 15
CYTOPLASM_MIN_SIZE = 15
CYTOPLASM_max_proj = False
CYTOPLASM_mean_proj = True
CYTOPLASM_select_slice = False
CYTOPLASM_selected_slice = 0

NUCLEUS_MIN_SIZE = 15
NUCLEUS_max_proj = False
NUCLEUS_mean_proj = True
NUCLEUS_select_slice = False
NUCLEUS_selected_slice = 0

#Detection
THRESHOLD = None
THRESHOLD_PENALTY = 1
DO_DENSE_REGIONS_DECONVOLUTION = False
DO_CLUSTER_COMPUTATION = False
DO_CLUSTER_COMPUTATION = False
SHOW_NAPARI_CORRECTOR = True
INTERACTIVE_THRESHOLD = False
VOXEL_SIZE = (1,2,3)

#Background removal
DO_BACKGROUND_REMOVAL = False
BACKGROUND_CHANNEL = 0

#Deconvolution
ALPHA = 0.5
BETA = 1.
GAMMA = 3.

#Clustering
CLUSTER_SIZE = 400
MIN_NUMBER_SPOTS = 5

#Coloc
COLOC_RANGE = 400

#Spots Extraction
DO_CSV = False
DO_EXCEL = False
SPOT_EXTRACTION_FOLDER = os.getcwd()


def get_default_settings() :
    return {
        "working_directory" : os.getcwd() if WORKING_DIRECTORY is None else WORKING_DIRECTORY,
        "do_background_removal" : DO_BACKGROUND_REMOVAL,
        "background_channel" : BACKGROUND_CHANNEL,
        "multichannel_stack" : IS_MULTICHANNEL,
        "stack_3D" : IS_3D_STACK,
        "detection_channel" : CHANNEL,
        "nucleus_channel" : NUC_CHANNEL,
        "flow_threshold" : FLOW_THRESHOLD,
        "cellprob_threshold" : CELLPROB_THRESHOD,
        "cytoplasm_diameter" : CYTOPLASM_DIAMETER,
        "nucleus_diameter" : NUC_DIAMETER,
        "anisotropy" : ANISOTROPY,
        "cytoplasm_model" : CYTO_MODEL,
        "nucleus_model" : NUC_MODEL,
        "show_segmentation" : SHOW_SEGMENTATION,
        "segment_only_nuclei" : SEGMENT_ONLY_NUCLEI,
        "do_3D_segmentation" : DO_3D_SEMGENTATION,
        "save_segmentation_visuals" : SAVE_SEGMENTATION_VISUAL,
        "threshold" : THRESHOLD,
        "threshold_penalty" : THRESHOLD_PENALTY,
        "do_dense_regions_deconvolution" : DO_DENSE_REGIONS_DECONVOLUTION,
        "do_cluster" : DO_CLUSTER_COMPUTATION,
        "show_napari_corrector" : SHOW_NAPARI_CORRECTOR,
        "interactive_threshold_selector" : INTERACTIVE_THRESHOLD,
        "alpha" : ALPHA,
        "beta" : BETA,
        "gamma" : GAMMA,
        "cluster_size" : CLUSTER_SIZE ,
        "min_spot" : MIN_NUMBER_SPOTS,
        "coloc_range" : COLOC_RANGE,
        "do_csv" : DO_CSV,
        "do_excel" : DO_EXCEL,
        "spot_extraction_folder" : SPOT_EXTRACTION_FOLDER,
        "voxel_size" : VOXEL_SIZE,
        "nucleus_min_size" : NUCLEUS_MIN_SIZE,
        "cytoplasm_min_size" : CYTOPLASM_MIN_SIZE,
        "cytoplasm_max_proj" : CYTOPLASM_max_proj,
        "cytoplasm_mean_proj" : CYTOPLASM_mean_proj,
        "cytoplasm_select_slice" : CYTOPLASM_select_slice,
        "cytoplasm_selected_slice" : CYTOPLASM_selected_slice,
        "nucleus_max_proj" : NUCLEUS_max_proj,
        "nucleus_mean_proj" : NUCLEUS_mean_proj,
        "nucleus_select_slice" : NUCLEUS_select_slice,
        "nucleus_selected_slice" : NUCLEUS_selected_slice,
    }


"""
Submodule to handle Small Fish settings.
"""

import os
import json
from .default_settings import get_default_settings
from pydantic import BaseModel, ValidationError
from typing import Tuple, Optional

class SettingsDict(BaseModel) :
    working_directory : str
    do_background_removal : bool
    background_channel : int
    multichannel_stack : bool
    stack_3D : bool
    detection_channel : int
    nucleus_channel : int
    anisotropy : float
    flow_threshold : float
    cellprob_threshold : float
    cytoplasm_diameter : int
    cytoplasm_min_size : int
    cytoplasm_max_proj : bool
    cytoplasm_mean_proj : bool
    cytoplasm_select_slice : bool
    cytoplasm_selected_slice : int
    nucleus_diameter : int
    cytoplasm_model : str
    nucleus_model : str
    nucleus_min_size : int
    nucleus_max_proj : bool
    nucleus_mean_proj : bool
    nucleus_select_slice : bool
    nucleus_selected_slice : int
    show_segmentation : bool
    segment_only_nuclei : bool
    do_3D_segmentation : bool
    save_segmentation_visuals : bool
    threshold : Optional[int]
    threshold_penalty : float
    do_dense_regions_deconvolution : bool
    do_cluster : bool
    show_napari_corrector : bool
    interactive_threshold_selector : bool
    alpha : float
    beta : float
    gamma : float
    cluster_size : int
    min_spot : int
    coloc_range : int
    do_csv : bool
    do_excel : bool
    spot_extraction_folder : str
    voxel_size : tuple
    


def get_settings() -> SettingsDict :

    setting_path = get_settings_path()

    if os.path.isfile(setting_path) :
        return _load_settings()
    else :
        settings = _init_settings()
        write_settings(settings)
        return settings

def _load_settings() :
    settings_path = get_settings_path()
    with open(settings_path, "r") as f:
        settings = json.load(f)
    
    try : settings = SettingsDict(**settings)

    except ValidationError as e :
        print(f"Incorrect settings, using default settings \n{e}")
        settings = _init_settings()
    
    return settings

def _init_settings() :
    default_settings = get_default_settings()
    return SettingsDict(**default_settings)

def get_settings_path() :
    return os.path.join(os.path.dirname(__file__) , "settings.json")

def write_settings(settings : SettingsDict) :
    if not isinstance(settings, SettingsDict) :
        raise TypeError("Expected SettingsDict type, got {}".format(type(settings)))
    else :
        settings_path = get_settings_path()
        with open(settings_path, mode="w") as f:
             json.dump(settings.dict(), f, indent=4)


"""
Contains cellpose wrappers to segmentate images.
"""

from cellpose.core import use_gpu
from skimage.measure import label
from ..hints import pipeline_parameters
from ..interface import get_settings
from ..gui import prompt, ask_cancel_segmentation, segmentation_prompt
from ..gui.napari_visualiser import show_segmentation as napari_show_segmentation
from ..interface import open_image, SettingsDict, get_settings

from .utils import from_label_get_centeroidscoords
from ._preprocess import ask_input_parameters
from ._preprocess import map_channels, reorder_shape, reorder_image_stack

from matplotlib.colors import ListedColormap
import matplotlib as mpl
import cellpose.models as models
import numpy as np
import bigfish.multistack as multistack
import bigfish.stack as stack
import bigfish.plot as plot
import FreeSimpleGUI as sg
import matplotlib.pyplot as plt
import os
from .utils import using_mps

def launch_segmentation(user_parameters: pipeline_parameters, nucleus_label, cytoplasm_label, batch_mode=False) :
    """
    Ask user for necessary parameters and perform cell segmentation (cytoplasm + nucleus) with cellpose.

    Input
    -----
    Image : np.ndarray[c,z,y,x]
        Image to use for segmentation.

    Returns
    -------
        cytoplasm_label, nucleus_label, user_parameters
    """

    segmentation_parameters : pipeline_parameters = user_parameters.copy()
    default = get_settings()

    #Ask for image parameters
    new_parameters = ask_input_parameters(user_parameters, ask_for_segmentation= True) #The image is open and stored inside user_parameters
    if type(new_parameters) == type(None) : #if user clicks 'Cancel'
        return nucleus_label , cytoplasm_label, user_parameters
    else :
        segmentation_parameters.update(new_parameters)

    map_ = map_channels(segmentation_parameters)
    if type(map_) == type(None) : #User clicks Cancel 
        return nucleus_label, cytoplasm_label, user_parameters
    segmentation_parameters['reordered_shape'] = reorder_shape(segmentation_parameters['shape'], map_)
    image = reorder_image_stack(map_, segmentation_parameters['image'])

    is_multichannel = segmentation_parameters["is_multichannel"]
    is_3D_stack = segmentation_parameters["is_3D_stack"]

    while True : # Loop if show_segmentation 
        #Default parameters
        path = os.getcwd()
        available_channels = list(range(image.shape[0]))
        available_slices = list(range(image.shape[is_multichannel])) #0 if not multichannel else 1

    #Ask user for parameters
    #if incorrect parameters --> set relaunch to True
        while True :
            segmentation_parameters.setdefault("other_nucleus_image", "")
            segmentation_parameters.setdefault("cytoplasm_channel", segmentation_parameters["detection_channel"])
            segmentation_parameters.setdefault("cytoplasm_segmentation_3D", segmentation_parameters["do_3D_segmentation"])
            segmentation_parameters.setdefault("nucleus_segmentation_3D", segmentation_parameters["do_3D_segmentation"])

            event, values = segmentation_prompt(
                saving_path= segmentation_parameters.setdefault("seg_control_saving_path", segmentation_parameters["working_directory"]),
                **segmentation_parameters
                )

            if event == 'Cancel' :
                cancel_segmentation = ask_cancel_segmentation()

                if cancel_segmentation :
                    return None, None, user_parameters
                else : 
                    continue

            #Extract parameters
            values, relaunch = _check_integrity_segmentation_parameters(values, segmentation_parameters, available_channels=available_channels, available_slices=available_slices)
            segmentation_parameters.update(values)
            if not relaunch : break

        if not batch_mode :
            #Launching segmentation
            waiting_layout = [
                [sg.Text("Running segmentation...")]
            ]
            window = sg.Window(
                title= 'small_fish',
                layout= waiting_layout,
                grab_anywhere= True,
                no_titlebar= False
            )

            window.read(timeout= 30, close= False)

        try :
            cytoplasm_label, nucleus_label = cell_segmentation(
                image,
                channels=[segmentation_parameters["cytoplasm_channel"], segmentation_parameters["nucleus_channel"]],
                do_only_nuc=segmentation_parameters["segment_only_nuclei"],
                external_nucleus_image = segmentation_parameters["other_nucleus_image"],
                nucleus_3D_segmentation=segmentation_parameters["nucleus_segmentation_3D"],
                cyto_3D_segmentation=segmentation_parameters["cytoplasm_segmentation_3D"],
                **segmentation_parameters
                )

        finally  : 
            if not batch_mode : window.close()

        if segmentation_parameters["show_segmentation"] :
            
            nucleus_label, cytoplasm_label = napari_show_segmentation(
                nuc_image=image[segmentation_parameters["nucleus_channel"]] if type(segmentation_parameters["other_nucleus_image"]) == type(None) else segmentation_parameters["other_nucleus_image"],
                nuc_label= nucleus_label,
                cyto_image=image[segmentation_parameters["cytoplasm_channel"]],
                cyto_label=cytoplasm_label,
                anisotrpy=segmentation_parameters["anisotropy"],
            )

            if nucleus_label.ndim == 3 : nucleus_label = np.max(nucleus_label, axis=0)
            if cytoplasm_label.ndim == 3 : cytoplasm_label = np.max(cytoplasm_label, axis=0)

            layout = [
                [sg.Text("Proceed with current segmentation ?")],
                [sg.Button("Yes"), sg.Button("No")]
            ]
            
            event, values = prompt(layout=layout, add_ok_cancel=False, add_scrollbar=False)
            if event == "No" :
                continue

        if type(segmentation_parameters["seg_control_saving_path"]) != type(None) and segmentation_parameters["save_segmentation_visuals"]:
            
            #Get backgrounds
            nuc_proj = image[segmentation_parameters["nucleus_channel"]]
            im_proj = image[segmentation_parameters["cytoplasm_channel"]]
            if im_proj.ndim == 3 :
                im_proj = stack.mean_projection(im_proj)
            if nuc_proj.ndim == 3 :
                nuc_proj = stack.mean_projection(nuc_proj)
            if nucleus_label.ndim == 3 :
                nucleus_label_proj = np.max(segmentation_parameters["nucleus_channel"],axis=0)
            else : 
                nucleus_label_proj = nucleus_label
            if cytoplasm_label.ndim == 3 :
                cytoplasm_label_proj = np.max(cytoplasm_label,axis=0)
            else : 
                cytoplasm_label_proj = cytoplasm_label

            
            #Call plots
            if segmentation_parameters["save_segmentation_visuals"]:
                output_path = segmentation_parameters["seg_control_saving_path"] + '/' + segmentation_parameters["filename"]
                nuc_path = output_path + "_nucleus_segmentation"
                cyto_path = output_path + "_cytoplasm_segmentation"

                #Plots boundaries
                plot.plot_segmentation_boundary(nuc_proj, cytoplasm_label_proj, nucleus_label_proj, boundary_size=2, contrast=True, show=False, path_output=nuc_path, title= "Nucleus segmentation (blue)", remove_frame=True,)
                if not segmentation_parameters["segment_only_nuclei"] : 
                    plot.plot_segmentation_boundary(im_proj, cytoplasm_label_proj, nucleus_label_proj, boundary_size=2, contrast=True, show=False, path_output=cyto_path, title="Cytoplasm Segmentation (red)", remove_frame=True)
            
                #Plots cell labels
                plot_labels(
                    nucleus_label,
                    path_output=output_path + "_nucleus_label_map.png",
                    show=False
                    )
                if not segmentation_parameters["segment_only_nuclei"] : 
                    plot_labels(
                        cytoplasm_label_proj,
                        path_output=output_path + "_cytoplasm_label_map.png",
                        show=False
                        )


        if cytoplasm_label.max() == 0 : #No cell segmented
            layout = [
            [sg.Text("No cell segmented. Proceed anyway ?")],
            [sg.Button("Yes"), sg.Button("No")]
        ]
            event, _ = prompt(layout=layout, add_ok_cancel=False)
            if event == "Yes" :
                return None, None, user_parameters
        else :
            break

    user_parameters.update(segmentation_parameters)
    return nucleus_label, cytoplasm_label, user_parameters

def cell_segmentation(
        reordered_image, 
        cytoplasm_model_name, 
        nucleus_model_name, 
        channels, 
        cytoplasm_diameter, 
        nucleus_diameter,
        nucleus_3D_segmentation=False,
        cyto_3D_segmentation=False,
        anisotropy = 1,
        nucleus_flow_threshold = 0.4,
        cytoplasm_flow_threshold = 0.4,
        nucleus_cellprob_threshold = 0.,
        cytoplasm_cellprob_threshold = 0.,
        do_only_nuc=False,
        external_nucleus_image = None,
        **segmentation_parameters : pipeline_parameters
        ) :

    nuc_channel = channels[1]
    

    if type(external_nucleus_image) != type(None) :
        nuc = external_nucleus_image
    else :
        nuc = reordered_image[nuc_channel]

    if nuc.ndim >= 3 and not nucleus_3D_segmentation:

        if segmentation_parameters["nucleus_max_proj"] : nuc = np.max(nuc, axis=0)
        elif segmentation_parameters["nucleus_mean_proj"] : nuc = np.mean(nuc, axis=0)
        elif segmentation_parameters["nucleus_select_slice"] : nuc = nuc[segmentation_parameters["nucleus_selected_slice"]]
        else : raise AssertionError("No option found for 2D nucleus seg. Should be impossible as this error is raised after integrity checks")
    
    
    nuc_label = _segmentate_object(
        nuc, 
        nucleus_model_name, 
        nucleus_diameter, 
        do_3D=nucleus_3D_segmentation, 
        anisotropy=anisotropy,
        flow_threshold= nucleus_flow_threshold,
        cellprob_threshold=nucleus_cellprob_threshold,
        min_size=segmentation_parameters["nucleus_min_size"]
        )
    
    if not do_only_nuc : 
        cyto_channel = channels[0]
        nuc = reordered_image[nuc_channel] if type(external_nucleus_image) == type(None) else external_nucleus_image

        if reordered_image[cyto_channel].ndim >= 3 and not cyto_3D_segmentation:
            if segmentation_parameters["cytoplasm_max_proj"] : cyto = np.max(reordered_image[cyto_channel], axis=0)
            elif segmentation_parameters["cytoplasm_mean_proj"] : cyto = np.mean(reordered_image[cyto_channel], axis=0)
            elif segmentation_parameters["cytoplasm_select_slice"] : cyto = reordered_image[cyto_channel][segmentation_parameters["cytoplasm_selected_slice"]]
            else : raise AssertionError("No option found for 2D cytoplasm seg. Should be impossible as this error is raised after integrity checks")
        else : 
            cyto = reordered_image[cyto_channel]
        if nuc.ndim >= 3 and not cyto_3D_segmentation:
            if segmentation_parameters["cytoplasm_max_proj"] : nuc = np.max(nuc, axis=0)
            elif segmentation_parameters["cytoplasm_mean_proj"] : nuc = np.mean(nuc, axis=0)
            elif segmentation_parameters["cytoplasm_select_slice"] : nuc = nuc[segmentation_parameters["cytoplasm_selected_slice"]]
            else : raise AssertionError("No option found for 2D cytoplasm seg. Should be impossible as this error is raised after integrity checks")

        reordered_image = np.zeros(shape=(2,) + cyto.shape)
        reordered_image[0] = cyto
        reordered_image[1] = nuc
        source = list(range(reordered_image.ndim))
        dest = source[-1:] + source[:-1]
        reordered_image = np.moveaxis(reordered_image, source=range(reordered_image.ndim), destination= dest)

        cytoplasm_label = _segmentate_object(
            reordered_image, 
            cytoplasm_model_name, 
            cytoplasm_diameter, 
            do_3D=cyto_3D_segmentation, 
            anisotropy=anisotropy,
            flow_threshold=cytoplasm_flow_threshold,
            cellprob_threshold=cytoplasm_cellprob_threshold,
            min_size=segmentation_parameters["cytoplasm_min_size"]
            )

        if cytoplasm_label.ndim == 3 and nuc_label.ndim == 2 :
            nuc_label = np.repeat(nuc_label[np.newaxis], len(cytoplasm_label), axis= 0)
        if nuc_label.ndim == 3 and cytoplasm_label.ndim == 2 :
            cytoplasm_label = np.repeat(cytoplasm_label[np.newaxis], len(nuc_label), axis= 0)

        nuc_label, cytoplasm_label = multistack.match_nuc_cell(nuc_label=nuc_label, cell_label=cytoplasm_label, single_nuc=True, cell_alone=False)
    else :
        cytoplasm_label = nuc_label

    return cytoplasm_label, nuc_label

def _segmentate_object(
        im : np.ndarray, 
        model_name : str, 
        object_size_px : int, 
        do_3D = False, 
        anisotropy : float = 1.0,
        flow_threshold : float = 0.4,
        cellprob_threshold : float = 0, 
        min_size = 15 #Default cellpose
        ) :
    

    model = models.CellposeModel(
        gpu= use_gpu(),
        pretrained_model= model_name,
        use_bfloat16= not using_mps()
    )

    label, flow, style = model.eval(
        im,
        diameter= object_size_px,
        do_3D= do_3D,
        z_axis=0 if do_3D else None,
        channel_axis= im.ndim -1 if im.ndim == 3+ do_3D else None,
        anisotropy=anisotropy,
        flow_threshold=flow_threshold,
        cellprob_threshold=cellprob_threshold,
        min_size=min_size,
        )
    
    label = np.array(label, dtype= np.int64)
    if not do_3D : label = remove_disjoint(label) # Too much time consuming in 3D
    
    return label

def _cast_segmentation_parameters(values:dict) :

    cast_rules = {
        'cytoplasm_diameter' : int,
        'nucleus_diameter' : int,
        'cytoplasm_channel' : int,
        'nucleus_channel' : int,
        'anisotropy' : float,
        'cytoplasm_flow_threshold' : float,
        'cytoplasm_cellprob_threshold' : float,
        'nucleus_flow_threshold' : float,
        'nucleus_cellprob_threshold' : float,
        'cytoplasm_anisotropy' : float,
        'nucleus_anisotropy' : float,
        'cytoplasm_selected_slice' : int,
        'nucleus_selected_slice' : int,
        'cytoplasm_min_size' : int,
        'nucleus_min_size' : int,
    }    

    for key, constructor in cast_rules.items() :
        try :
            if key in values.keys() : values[key] = constructor(values[key])
        except ValueError :
            pass

    #None if default
    if values['cytoplasm_model_name'] == '' :
        values['cytoplasm_model_name'] = None

    if values['nucleus_model_name'] == '' :
        values['nucleus_model_name'] = None

    return values

def _check_integrity_segmentation_parameters(
    values : pipeline_parameters, 
    user_parameters : pipeline_parameters, 
    available_channels : list, 
    available_slices : list
    ) :
    """
    Check that parameters entered by user comply with what we need to do with them. Type checking is handle in _cast_segmentation_parameters, if any parameters could not be cast to 
    appropriate type, it is set to None.
    """

    values : pipeline_parameters = _cast_segmentation_parameters(values)
    is_3D_stack = user_parameters["is_3D_stack"]
    is_multichannel = user_parameters["is_multichannel"]

    relaunch= False
    #Control plots
    if values["save_segmentation_visuals"] :
        if not os.path.isdir(values["seg_control_saving_path"]) :
            relaunch=True
            sg.popup(f"{values["seg_control_saving_path"]} is not a directory.")
            values["seg_control_saving_path"] = user_parameters.get("seg_control_saving_path")
        if values["filename"] == "" :
            relaunch=True
            sg.popup("Enter a filename for control plots.")
            values["filename"] = user_parameters.get("filename")

    #2D/3D seg
    if is_3D_stack :
        for obj in ["cytoplasm","nucleus"] :
            assert values[obj + "_radio_2D"] + values[obj + "_radio_3D"] == 1, f"Incorrect number of segmentation dimension was selected (should be 1) for {obj}"
            assert values[obj + "_max_proj"] + values[obj + "_mean_proj"] + values[obj + "_select_slice"] == 1, f"Incorrect number of 2D segmentation options were selected (should be 1) for {obj}"

            if values[obj + "_radio_2D"] : values[obj + "_segmentation_3D"] = False
            else : values[obj + "_segmentation_3D"] = True

        #Selected slice
        for key in ['cytoplasm_selected_slice', 'nucleus_selected_slice'] :
            if not isinstance(values[key], int) :
                relaunch = True
                sg.popup("Invalid slice number for cytoplasm 2D segmentation.")
                values[key] = user_parameters[key]
            elif values[key] not in available_slices :
                relaunch = True
                sg.popup(f"Invalid slice number for cytoplasm 2D segmentation, selecgted between 0 and {available_slices[-1]}")
                values[key] = user_parameters[key]
        
        #anisotropy
        if not "anisotropy" in values.keys() :
            values["anisotropy"] = 1
        if not isinstance(values["nucleus_anisotropy"], (float, int)) :
            relaunch=True
            sg.popup("Invalid value for nucleus anisotropy, must be a positive float.")
            values["anisotropy"] = user_parameters.get("anisotropy")
        elif values["nucleus_anisotropy"] < 0 :
            relaunch=True
            sg.popup("Invalid value for nucleus anisotropy, must be a positive float.")
            values["anisotropy"] = user_parameters.get("anisotropy")

        if not values["segment_only_nuclei"] :
            if not isinstance(values["cytoplasm_anisotropy"], (float, int)) :
                relaunch=True
                sg.popup("Invalid value for cytoplasm anisotropy, must be a positive float.")
                values["anisotropy"] = user_parameters.get("anisotropy")
            elif values["cytoplasm_anisotropy"] < 0 :
                relaunch=True
                sg.popup("Invalid value for cytoplasm anisotropy, must be a positive float.")
                values["anisotropy"] = user_parameters.get("anisotropy")

        if values["segment_only_nuclei"] :
            values["anisotropy"] = values["nucleus_anisotropy"]
        elif isinstance(values["cytoplasm_anisotropy"], (float, int)) and isinstance(values["nucleus_anisotropy"], (float, int)) and values["nucleus_segmentation_3D"] and values["cytoplasm_segmentation_3D"]:
            if (not values["cytoplasm_anisotropy"] == values["nucleus_anisotropy"]) and (not values["segment_only_nuclei"]) :
                relaunch=True
                sg.popup("Anisotropy must be equal for nucleus and cytoplasm segmentation")
                values["anisotropy"] = user_parameters.get("anisotropy")
            else :
                values["anisotropy"] = values["nucleus_anisotropy"]
        elif isinstance(values["nucleus_anisotropy"], (float, int)) and values["nucleus_segmentation_3D"] :
            values["anisotropy"] = values["nucleus_anisotropy"]
        elif isinstance(values["cytoplasm_anisotropy"], (float, int)) and values["cytoplasm_segmentation_3D"] :
            values["anisotropy"] = values["cytoplasm_anisotropy"]

        if not isinstance(values["anisotropy"], (float,int)) :
            sg.popup("Anisotropy must be a positive float.")
            relaunch = True
            values['anisotropy'] = user_parameters.get("anisotropy")
    
    else :
        values["cytoplasm_segmentation_3D"] = False
        values["nucleus_segmentation_3D"] = False
        values["anisotropy"] = 1
        for obj in ["nucleus","cytoplasm"] :
            values[obj + "_max_proj"] = False
            values[obj + "_mean_proj"] = True
            values[obj + "_select_slice"] = False
            values[obj + "_selected_slice"] = 0

    #Min size
    if not isinstance(values["cytoplasm_min_size"], int) :
        relaunch=True
        sg.popup("Invalid value for cytoplasm min size parameter, must be a positive int.")
        values["cytoplasm_min_size"] = user_parameters["cytoplasm_min_size"]
    elif values["cytoplasm_min_size"] < 0 :
        relaunch=True
        sg.popup("Invalid value for cytoplasm min size parameter, must be a positive int.")
        values["cytoplasm_min_size"] = user_parameters["cytoplasm_min_size"]

    if not isinstance(values["nucleus_min_size"], int) :
        relaunch=True
        sg.popup("Invalid value for nucleus min size parameter, must be a positive int.")
        values["nucleus_min_size"] = user_parameters["nucleus_min_size"]
    elif values["nucleus_min_size"] < 0 :
        relaunch=True
        sg.popup("Invalid value for nucleus min size parameter, must be a positive int.")
        values["nucleus_min_size"] = user_parameters["nucleus_min_size"]

    #flow thresholds
    if type(values['nucleus_flow_threshold']) != float : 
        sg.popup('Invalid value for flow threshold in nuc parameters, must be a float between 0 and 1.')
        values['nucleus_flow_threshold'] = user_parameters['nucleus_flow_threshold']
        relaunch= True
    elif abs(values['nucleus_flow_threshold']) > 1 or values['nucleus_flow_threshold'] < 0 :
        sg.popup('Invalid value for flow threshold in cyto parameters, must be a float between 0 and 1.')
        values['nucleus_flow_threshold'] = user_parameters['nucleus_flow_threshold']
        relaunch= True 

    if type( values['cytoplasm_flow_threshold']) != float : 
        sg.popup('Invalid value for flow threshold in cyto parameters, must be a float between 0 and 1.')
        values['cytoplasm_flow_threshold'] = user_parameters['cytoplasm_flow_threshold']
        relaunch= True
    elif abs(values['cytoplasm_flow_threshold']) > 1 or values['cytoplasm_flow_threshold'] < 0 :
        sg.popup('Invalid value for flow threshold in cyto parameters, must be a float between 0 and 1.')
        values['cytoplasm_flow_threshold'] = user_parameters['cytoplasm_flow_threshold']
        relaunch= True 
    
    #cellprob thresholds
    if type(values['nucleus_cellprob_threshold']) != float : 
        sg.popup('Invalid value for cellprob threshold in nuc parameters, must be a float between -3 and +3.')
        values['nucleus_cellprob_threshold'] = user_parameters['nucleus_cellprob_threshold']
        relaunch= True
    elif abs(values['nucleus_cellprob_threshold']) > 3 :
        sg.popup('Invalid value for cellprob threshold in cyto parameters, must be a float between -3 and +3.')
        values['nucleus_cellprob_threshold'] = user_parameters['nucleus_cellprob_threshold']
        relaunch= True

    if type(values['cytoplasm_cellprob_threshold']) != float :
        sg.popup('Invalid value for cellprob threshold in cyto parameters, must be a float between -3 and +3.')
        values['cytoplasm_cellprob_threshold'] = user_parameters['cytoplasm_cellprob_threshold']
        relaunch= True
    elif abs(values['cytoplasm_cellprob_threshold']) > 3 :
        sg.popup('Invalid value for cellprob threshold in cyto parameters, must be a float between -3 and +3.')
        values['cytoplasm_cellprob_threshold'] = user_parameters['cytoplasm_cellprob_threshold']
        relaunch= True

    
    #Cytoplasm parameters
    if type(values["cytoplasm_model_name"]) != str  and not do_only_nuc:
        sg.popup('Invalid cytoplasm model name.')
        values['cytoplasm_model_name'] = user_parameters['cytoplasm_model_name']
        relaunch= True
    if is_multichannel :
        if values["cytoplasm_channel"] not in available_channels and not do_only_nuc:
            sg.popup('For given input image please select channel in {0}\ncytoplasm_channel : {1}'.format(available_channels, cytoplasm_channel))
            relaunch= True
            values['cytoplasm_channel'] = user_parameters['cytoplasm_channel']
    else :
        values["cytoplasm_channel"] = ...

    if type(values["cytoplasm_diameter"]) not in [int, float] and not do_only_nuc:
        sg.popup("Incorrect cytoplasm size.")
        relaunch= True
        values['cytoplasm_diameter'] = user_parameters['cytoplasm_diameter']

    #Nucleus parameters
    if type(values["nucleus_model_name"]) != str :
        sg.popup('Invalid nucleus model name.')
        values['nucleus_model_name'] = user_parameters['nucleus_model_name']
        relaunch= True
    
    if is_multichannel :
        if values["nucleus_channel"] not in available_channels :
            sg.popup('For given input image please select channel in {0}\nnucleus channel : {1}'.format(available_channels, nucleus_channel))
            relaunch= True
            values['nucleus_channel'] = user_parameters['nucleus_channel']
    else : 
        values["nucleus_channel"] = ...

    if type(values["nucleus_diameter"]) not in [int, float] :
        sg.popup("Incorrect nucleus size.")
        relaunch= True
        values['nucleus_diameter'] = user_parameters['nucleus_diameter']

    if values["other_nucleus_image"] != '' :
        if os.path.isdir(values["other_nucleus_image"]) :
            values['other_nucleus_image'] = None
        elif not os.path.isfile(values["other_nucleus_image"]) :
            sg.popup("Nucleus image is not a file.")
            relaunch=True
            values['other_nucleus_image'] = None
        else :
            try :
                nucleus_image = open_image(other_nucleus_image)
            except Exception as e :
                sg.popup("Could not open image.\n{0}".format(e))
                relaunch=True
                values['other_nucleus_image'] = user_parameters['other_nucleus_image']
            else :
                if nucleus_image.ndim != image.ndim - is_multichannel :
                    sg.popup("Nucleus image dimension missmatched. Expected same dimension as cytoplasm_image for monochannel or same dimension as cytoplasm_image -1 for is_multichannel\ncytoplasm dimension : {0}, nucleus dimension : {1}".format(image.ndim, nucleus_image.ndim))
                    nucleus_image = None
                    relaunch=True
                    values['other_nucleus_image'] = user_parameters['other_nucleus_image']
                
                elif nucleus_image.shape != image[cytoplasm_channel].shape :
                    sg.popup("Nucleus image shape missmatched. Expected same shape as cytoplasm_image \ncytoplasm shape : {0}, nucleus shape : {1}".format(image[cytoplasm_channel].shape, nucleus_image.shape))
                    nucleus_image = None
                    relaunch=True
                    values['other_nucleus_image'] = user_parameters['other_nucleus_image']

    else :
        values["other_nucleus_image"] = None

    return values, relaunch


def remove_disjoint(image):
    """
    *CODE FROM BIG-FISH (LICENCE IN __INIT__.PY) IMPORTED TO AVOID IMPORT ERROR WHEN BIGFISH SEGMENTATION MODULE INITIALISES : try to import deprecated module for python 3.8

    For each instances with disconnected parts, keep the larger one.

    Parameters
    ----------
    image : np.ndarray, np.int, np.uint or bool
        Labelled image with shape (z, y, x) or (y, x).

    Returns
    -------
    image_cleaned : np.ndarray, np.int or np.uint
        Cleaned image with shape (z, y, x) or (y, x).

    """
    # check parameters
    stack.check_array(
        image,
        ndim=[2, 3],
        dtype=[np.uint8, np.uint16, np.int32, np.int64, bool])

    # handle boolean array
    cast_to_bool = False
    if image.dtype == bool:
        cast_to_bool = bool
        image = image.astype(np.uint8)

    # initialize cleaned labels
    image_cleaned = np.zeros_like(image)

    # loop over instances
    max_label = image.max()
    for i in range(1, max_label + 1):

        # get instance mask
        mask = image == i

        # check if an instance is labelled with this value
        if mask.sum() == 0:
            continue

        # get an index for each disconnected part of the instance
        labelled_mask = label(mask)
        indices = sorted(list(set(labelled_mask.ravel())))
        if 0 in indices:
            indices = indices[1:]

        # keep the largest part of the instance
        max_area = 0
        mask_instance = None
        for j in indices:
            mask_part_j = labelled_mask == j
            area_j = mask_part_j.sum()
            if area_j > max_area:
                max_area = area_j
                mask_instance = mask_part_j

        # add instance in the final label
        image_cleaned[mask_instance] = i

    if cast_to_bool:
        image_cleaned = image_cleaned.astype(bool)

    return image_cleaned

def plot_segmentation(
        cyto_image : np.ndarray, 
        cyto_label : np.ndarray, 
        nuc_image : np.ndarray, 
        nuc_label : np.ndarray,
        path :str, 
        do_only_nuc=False
        ) :

    if nuc_image.ndim == 3 :
        nuc_image = np.max(nuc_image,axis=0)
    
    if cyto_label.ndim == 3 :
        cyto_label = np.max(cyto_label,axis=0)
    
    if nuc_label.ndim == 3 :
        nuc_label = np.max(nuc_label,axis=0)
    
    if cyto_image.ndim == 3 :
        cyto_image = np.max(cyto_image,axis=0)
    
    plot.plot_segmentation_boundary(
        image=nuc_image,
        nuc_label= nuc_label,
        boundary_size= 3,
        contrast=True,
        path_output=path + "_nuclei_segmentation.png",
        show=False,
    )


    if not do_only_nuc :
        if cyto_image.ndim == 3 :
            cyto_image = np.max(cyto_image,axis=0)
    
        plot.plot_segmentation_boundary(
            image=cyto_image,
            cell_label= cyto_label,
            nuc_label= nuc_label,
            boundary_size= 3,
            contrast=True,
            path_output=path + "_cytoplasm_segmentation.png",
            show=False,
        )

def plot_labels(labelled_image: np.ndarray, path_output:str = None, show= True, axis= False, close= True):
    """
    Plot a labelled image and indicate the label number at the center of each region.
    """
    stack.check_parameter(labelled_image = (np.ndarray, list), show = (bool))
    if isinstance(labelled_image, np.ndarray) : 
        stack.check_array(labelled_image, ndim= 2)
        labelled_image = [labelled_image]
    
    #Setting a colormap with background to white so all cells can be visible
    viridis = mpl.colormaps['viridis'].resampled(256)
    newcolors = viridis(np.linspace(0, 1, 256))
    white = np.array([1, 1, 1, 1])
    newcolors[0, :] = white
    newcmp = ListedColormap(newcolors)

    plt.figure(figsize= (10,10))
    rescaled_image = stack.rescale(np.array(labelled_image[0], dtype= np.int32), channel_to_stretch= 0)
    rescaled_image[rescaled_image == 0] = -100
    plot = plt.imshow(rescaled_image, cmap=newcmp)
    plot.axes.get_xaxis().set_visible(axis)
    plot.axes.get_yaxis().set_visible(axis)
    plt.tight_layout()

    for index in range(0, len(labelled_image)) :
        centroid_dict = from_label_get_centeroidscoords(labelled_image[index])
        labels = centroid_dict["label"]
        Y = centroid_dict["centroid-0"]
        X = centroid_dict["centroid-1"]
        centroids = zip(Y,X)

        for label in labels :
            y,x = next(centroids)
            y,x = round(y), round(x)
            an = plt.annotate(str(label), [round(x), round(y)])

    if not axis : plt.cla
    if show : plt.show()
    if path_output != None :
        stack.check_parameter(path_output = (str))
        plt.savefig(path_output)
    if close : plt.close()

    return plot


"""
Sub-module to handle individual spot extraction.

"""

import numpy as np
import pandas as pd
from ..interface.inoutput import write_results

def launch_spots_extraction(
        acquisition_id,
        user_parameters,
        image,
        spots,
        cluster_id,
        nucleus_label,
        cell_label,
) :
    Spots = compute_Spots(
        acquisition_id=acquisition_id,
        image=image,
        spots=spots,
        cluster_id= cluster_id,
        nucleus_label=nucleus_label,
        cell_label=cell_label,
    )

    did_output = write_results(
        Spots,
        path= user_parameters['spots_extraction_folder'],
        filename= user_parameters['spots_filename'],
        do_excel=user_parameters['do_spots_excel'],
        do_csv=user_parameters['do_spots_csv'],
        )
    
    if did_output : print("Individual spots extracted at {0}".format(user_parameters['spots_extraction_folder']))

def compute_Spots(
        acquisition_id : int,
        image : np.ndarray,
        spots : np.ndarray,
        cluster_id : np.ndarray,
        nucleus_label : np.ndarray = None,
        cell_label : np.ndarray = None,
) :

    if len(spots) == 0 :
        return pd.DataFrame()

    if type(cluster_id) == type(None) : #When user doesn't select cluster
        cluster_id = [np.nan]*len(spots)

    index = list(zip(*spots))
    index = tuple(index)
    spot_intensities_list = list(image[index])
    if type(nucleus_label) != type(None) :
        if nucleus_label.ndim == 2 :
            in_nuc_list = list(nucleus_label.astype(bool)[index[-2:]]) #Only plane coordinates
        else :
            in_nuc_list = list(nucleus_label.astype(bool)[index])
    else :
        in_nuc_list = np.nan
    if type(cell_label) != type(None) :
        if cell_label.ndim == 3 :
            cell_label_list = list(cell_label[index])
        else :    
            cell_label_list = list(cell_label[index[-2:]]) #Only plane coordinates
    else :
        cell_label_list = np.nan
    id_list = np.arange(len(spots))

    coord_list = list(zip(*index))

    Spots = pd.DataFrame({
        'acquisition_id' : [acquisition_id] * len(spots),
        'spot_id' : id_list,
        'intensity' : spot_intensities_list,
        'cell_label' : cell_label_list,
        'in_nucleus' : in_nuc_list,
        'coordinates' : coord_list,
        'cluster_id' : cluster_id,
    })

    return Spots
    
def load_spots(
        table_path : str
        ) -> pd.DataFrame :
    
    if table_path.endswith('.csv') :
        Spots = pd.read_csv(table_path, sep= ";")
    elif table_path.endswith('.xlsx') or table_path.endswith('.xls') :
        Spots = pd.read_excel(table_path)
    elif table_path.endswith('.feather') :
        Spots = pd.read_feather(table_path)
    else :
        raise ValueError("Table format not recognized. Please use .csv, .xlsx or .feather files.")
    
    if "coordinates" in Spots.columns :
        pass
    elif "y" in Spots.columns and "x" in Spots.columns :
        if "z" in Spots.columns :
            pass
        else :
            pass
    else :
        raise ValueError("Coordinates information not found in table. Please provide a 'coordinates' column with tuples (z,y,x) or (y,x) or 'y' and 'x' columns.")

    return Spots

def reconstruct_acquisition_data(
        Spots : pd.DataFrame,
        max_id : int,
        filename : str,
        voxel_size : 'tuple[int]'
        ) -> pd.DataFrame:
    """
    Aim : creating a acquisition to add to result_dataframe from loaded spots for co-localization use  

    **Needed keys for colocalization**
        * acquisition_id
        * name
        * spots : np.ndarray[int] (nb_spots, nb_coordinates)
        * clusters : np.ndarray[int] (nb_cluster, nb_coordinate + 2)
        * spots_cluster_id : list[int]
        * voxel_size : tuple[int]
        * shape : tuple[int]
        * filename : str
    """
    
    spots = reconstruct_spots(Spots['coordinates'])
    has_clusters = not Spots['cluster_id'].isna().all()
    spot_number = len(spots)

    if has_clusters :

        clusters = np.empty(shape=(0,5), dtype=int) #useless for coloc only needded in columns to enable coloc on clusters
        spot_cluster_id = Spots['cluster_id'].to_numpy().astype(int).tolist()

        new_acquisition = pd.DataFrame({
            'acquisition_id' : [max_id + 1],
            'name' : ["loaded_spots_{}".format(max_id + 1)],
            'threshold' : [0],
            'spots' : [spots],
            'clusters' : [clusters],
            'spots_cluster_id' : [spot_cluster_id],
            'spot_number' : [spot_number],
            'filename' : [filename],
            'voxel_size' : [voxel_size],
        })
    else :
        new_acquisition = pd.DataFrame({
            'acquisition_id' : [max_id + 1],
            'name' : ["loaded_spots_{}".format(max_id + 1)],
            'threshold' : [0],
            'spots' : [spots],
            'spot_number' : [spot_number],
            'filename' : [filename],
            'voxel_size' : [voxel_size],
        })

    return new_acquisition

def reconstruct_spots(
        coordinates_serie : pd.Series
        ) :
    spots = coordinates_serie.str.replace('(','').str.replace(')','')
    spots = spots.str.split(',')
    spots = spots.apply(np.array)
    spots = np.array(spots.to_list()).astype(int)

    return spots


def reconstruct_cell_data(
        Spots : pd.DataFrame,
        max_id : int,
        ) :
    
    has_cluster = not Spots['cluster_id'].isna().all()
    coordinates = reconstruct_spots(Spots['coordinates'])
    Spots['coordinates'] = pd.Series(coordinates.tolist(), dtype=object, index= Spots.index)

    cell = Spots.groupby('cell_label')['coordinates'].apply(np.array).rename("rna_coords").reset_index(drop=False)
    
    #Handle cells with no spots
    na_mask =cell[cell['rna_coords'].isna()].index
    cell.loc[na_mask, ['rna_coords']] = pd.Series([np.empty(shape=(0,3))]*len(na_mask), dtype= object, index=na_mask)
    
    cell['total_rna_number'] = cell['rna_coords'].apply(len)
    
    if has_cluster :
        cell['clustered_spots_coords'] = Spots[Spots['cluster_id'] !=-1].groupby('cell_label')['coordinates'].apply(np.array).rename("clustered_spots_coords")
        
        #Handle cells with no clusters
        na_mask =cell[cell['clustered_spots_coords'].isna()].index
        cell.loc[na_mask, ['clustered_spots_coords']] = pd.Series([np.empty(shape=(0,3))]*len(na_mask), dtype= object, index=na_mask)
        
        cell['clustered_spot_number'] = cell['clustered_spots_coords'].apply(len)

    cell['acquisition_id'] = max_id + 1
    cell['name'] = "loaded_spots_{}".format(max_id + 1)
    cell = cell.rename(columns={"cell_label": "cell_id"})

    return cell


import numpy as np
import os

def output_masks(
        batch_path : str,
        acquisition_name : str,
        nucleus_label : np.ndarray,
        cytoplasm_label : np.ndarray = None,
) :

    os.makedirs(batch_path + "/segmentation_masks/", exist_ok=True)
    output_path = batch_path + "/segmentation_masks/{0}".format(acquisition_name)

    np.save(output_path + "_nucleus.npy", arr= nucleus_label)
    if type(cytoplasm_label) != type(None) :
        np.save(output_path + "_cytoplasm.npy", arr= cytoplasm_label)




CELLPROB_TOOLTIP = """ (float) from -6. to +6.
The network predicts 3 outputs: flows in X, flows in Y, and cell “probability”. 
The predictions the network makes of the probability are the inputs to a sigmoid centered at zero (1 / (1 + e^-x)), so they vary from around -6 to +6. 
The pixels greater than the cellprob_threshold are used to run dynamics and determine ROIs. 
The default is cellprob_threshold=0.0. 
Decrease this threshold if cellpose is not returning as many ROIs as you’d expect. 
Similarly, increase this threshold if cellpose is returning too ROIs particularly from dim areas.
""" #Cellpose 4.0.6 doc

FLOW_THRESHOLD_TOOLTIP = """ (float) from 0. to 1.
The flow_threshold parameter is the maximum allowed error of the flows for each mask. 
Increase this threshold if cellpose is not returning as many ROIs as you’d expect. 
Similarly, decrease this threshold if cellpose is returning too many ill-shaped ROIs.
""" #Cellpose 4.0.6 doc

MIN_SIZE_TOOLTIP = """ (int) >0 ; pixel
Default to 15px. Minimum size for predicted regions during cytoplasm or nucleus segmentation. Smaller regions will be discarded.
"""

REMOVE_BACKGROUND_TOOLTIP = """RANSAC fit + substraction method is applied using selected channel as background and detection channel as signal. \nResulting signal is passed for spot detection and quantification but NOT for segmentation."""


"""
Submodule containing custom class for napari widgets
"""
import os
import numpy as np
import pandas as pd
import bigfish.detection as detection
import bigfish.stack as stack
from skimage.segmentation import find_boundaries
from ..pipeline._bigfish_wrapers import _apply_log_filter, _local_maxima_mask

from napari.layers import Labels, Points, Image
from napari.utils.events import Event, EmitterGroup
from magicgui import magicgui
from magicgui.widgets import SpinBox, Container
from bigfish.detection import spots_thresholding, automated_threshold_setting
from napari.types import LayerDataTuple

from abc import ABC, abstractmethod
from typing import Tuple, List
from ..utils import compute_anisotropy_coef

from AF_eraser import remove_autofluorescence_RANSACfit
from pathlib import Path
from ..interface import open_image


class NapariWidget(ABC) :
    """
    Common super class for custom widgets added to napari interface during run
    Each sub class as a specific function, but the widget can be acess with attribute .widget
    """
    def __init__(self):
        self.widget = self._create_widget()

    @abstractmethod
    def _create_widget(self) :
        """
        This should return a widget you can add to the napari (QWidget)
        """
        pass

# Corrector widgets

class ClusterWidget(NapariWidget) :
    """
    Widget for clusters interaction are all init with cluster_layer and single_layer
    """
    def __init__(self, cluster_layer : Points, single_layer : Points):
        self.cluster_layer = cluster_layer
        self.single_layer = single_layer
        super().__init__()

class ClusterWizard(ABC) :
    """
    Common super class for all classes that will interact on single layer and cluster layer to synchronise them or modify their display.
    Their action is started through 'start_listening' method.
    To register them in CLUSTER_WIZARD they should only take single_layer and cluster_layer as arguments
    """

    def __init__(self, single_layer : Points, cluster_layer : Points):
        self.single_layer = single_layer
        self.cluster_layer = cluster_layer
        self.start_listening()
    
    def start_listening(self) :
        """
        This activate the class function. Returns None
        """
        pass


CLUSTER_WIZARDS = []
def register_cluster_wizard(cls):
    """
    Helper to register all clusters related wizard class. Object to be instanciated upon launching napari but that have no widget.
    """
    CLUSTER_WIZARDS.append(cls)
    return cls

def initialize_all_cluster_wizards(single_layer: Points, cluster_layer: Points):
    """
    Initialize all wizards for cluster interaction in Napari
    """
    return [
        cls(single_layer, cluster_layer)
        for cls in CLUSTER_WIZARDS
    ]


class CellLabelEraser(NapariWidget) :
    """
    Widget for deleting cells from multiple label layers in a Napari viewer.
    """
    def __init__(self, label_list: 'list[Labels]'):
        self.label_list = label_list
        if len(self.label_list) == 0 : raise ValueError("Empty label list")
        for label_layer in self.label_list :
            label_layer.events.selected_label.connect((self, 'update'))
        super().__init__()

    def update(self, event) :
        layer : Labels = event.source
        new_label = layer.selected_label
        self.widget.label_number.value = new_label
        self.widget.update()
    
    def _create_widget(self) :
        @magicgui(
                call_button="Delete cell",
                auto_call=False
                )
        def label_eraser(label_number: int) -> None :

            for i, label in enumerate(self.label_list) :
                self.label_list[i].data[label.data == label_number] = 0
                label.refresh()

        return label_eraser


class FreeLabelPicker(NapariWidget) :
    """
    This widget gives the user a free label number
    """
    def __init__(self, label_list : 'list[Labels]'):
        self.label_list = label_list
        if len(self.label_list) == 0 : raise ValueError("Empty label list")
        super().__init__()
    
    def _create_widget(self) :
        @magicgui(
            call_button="Pick free label",
            auto_call=False
        )
        def label_pick()->None :
            max_list = [label_layer.data.max() for label_layer in self.label_list]
            new_label = max(max_list) + 1
            for label_layer in self.label_list :
                label_layer.selected_label = new_label
                label_layer.refresh()

        return label_pick


class SegmentationReseter(NapariWidget) :
    """
    This widget reset the segmentation mask as it used to be when iniating the instance
    """
    def __init__(self, label_list: 'list[Labels]'):
        self.label_list = label_list
        if len(self.label_list) == 0 : raise ValueError("Empty label list")
        self.save = self._get_save(label_list)
        super().__init__()
        
    
    def _get_save(self, label_list : 'list[Labels]') :
        return [label.data.copy() for label in label_list]

    def _create_widget(self) :
        @magicgui(
            call_button= 'Reset segmentation',
            auto_call=False,
        )
        def reset_segmentation() -> None:
            for save_data, layer in zip(self.save, self.label_list) :
                layer.data = save_data.copy()
                layer.refresh()

        return reset_segmentation

class ChangesPropagater(NapariWidget) :
    """
    Apply the changes across the vertical direction (Zstack) if confling values are found for a pixel, max label is kept.
    """
    def __init__(self, label_list):
        self.label_list = label_list
        for label_layer in self.label_list :
            label_layer.events.selected_label.connect((self, 'update'))
        super().__init__()

    def _create_widget(self) :
        @magicgui(
            call_button='Expand label',
            auto_call=False,
        )
        def apply_changes(label_number : int) -> None:
            for layer in self.label_list :
                slices = layer.data.shape[0]
                masked_layer = layer.data.copy()
                masked_layer[masked_layer != label_number] = 0
                masked_layer = np.max(masked_layer, axis=0)
                expanded_layer = np.repeat(masked_layer[np.newaxis], slices, axis=0)
                layer.data[expanded_layer == label_number] = expanded_layer[expanded_layer == label_number]
                layer.refresh()
        return apply_changes

    def update(self, event) :
        layer : Labels = event.source
        new_label = layer.selected_label
        self.widget.label_number.value = new_label
        self.widget.update()

class ClusterIDSetter(ClusterWidget) :
    """
    Allow user to set selected single spots to chosen cluster_id
    """
    def __init__(self, single_layer : Points, cluster_layer : Points):
        super().__init__(cluster_layer, single_layer)

    def _create_widget(self):

        @magicgui(
                call_button= "Set cluster ID",
                auto_call= False,
                cluster_id= {'min' : -1},
        )
        def set_cluster_id(cluster_id : int) :
            if cluster_id == -1 or cluster_id in self.cluster_layer.features['cluster_id'] :
                spots_selection = list(self.single_layer.selected_data)
                cluster_id_in_selection = list(self.single_layer.features.loc[spots_selection,["cluster_id"]].to_numpy().flatten()) + [cluster_id]
                self.single_layer.features.loc[spots_selection,["cluster_id"]] = cluster_id

                for cluster_id in np.unique(cluster_id_in_selection): # Then update number of spots in cluster
                    if cluster_id == -1 : continue
                    new_spot_number = len(self.single_layer.features.loc[self.single_layer.features['cluster_id'] == cluster_id])
                    self.cluster_layer.features.loc[self.cluster_layer.features['cluster_id'] == cluster_id, ["spot_number"]] = new_spot_number
                self.cluster_layer.events.features()
            else :
                print(f"Not cluster with id {cluster_id} was found.")

            self.cluster_layer.selected_data.clear()

        return set_cluster_id

class ClusterMerger(ClusterWidget) :
    """
    Merge all selected clusters by replacing cluster ids of all clusters and belonging points with min for cluster id.
    """
    def __init__(self, cluster_layer, single_layer):
        super().__init__(cluster_layer, single_layer)
    
    
    def _create_widget(self):

        @magicgui(
            call_button="Merge Clusters",
            auto_call=False
        )
        def merge_cluster()-> None :
            selected_clusters = list(self.cluster_layer.selected_data)
            if len(selected_clusters) == 0 : return None

            selected_cluster_ids = self.cluster_layer.features.loc[selected_clusters,['cluster_id']].to_numpy().flatten()
            new_cluster_id = selected_cluster_ids.min()

            #Dropping selected clusters
            self.cluster_layer.data = np.delete(self.cluster_layer.data, selected_clusters, axis=0)
            self.cluster_layer.features = self.cluster_layer.features.drop(selected_clusters, axis=0)

            #Updating spots
            belonging_spots = self.single_layer.features.loc[self.single_layer.features['cluster_id'].isin(selected_cluster_ids)].index
            self.single_layer.features.loc[belonging_spots, ["cluster_id"]] = new_cluster_id

            #Creating new cluster
            centroid = list(self.single_layer.data[belonging_spots].mean(axis=0).round().astype(int))
            spot_number = len(belonging_spots)
            self.cluster_layer.data = np.append(
                self.cluster_layer.data,
                [centroid],
                axis=0
            )

            last_index = len(self.cluster_layer.data) - 1
            self.cluster_layer.features.loc[last_index, ['cluster_id']] = new_cluster_id
            self.cluster_layer.features.loc[last_index, ['spot_number']] = spot_number

            self.cluster_layer.selected_data.clear()
            self.cluster_layer.refresh()

        return merge_cluster

class ClusterUpdater(NapariWidget) :
    """
    Relaunch clustering algorithm taking into consideration new spots, new clusters and deleted clusters.
    """
    def __init__(
            self, 
            single_layer : Points, 
            cluster_layer : Points, 
            default_cluster_radius : int, 
            default_min_spot : int,
            voxel_size : 'tuple[int]'
            ):
        self.single_layer = single_layer
        self.cluster_layer = cluster_layer
        self.cluster_radius = default_cluster_radius
        self.min_spot = default_min_spot
        self.voxel_size = voxel_size
        super().__init__()

    def _compute_clusters(
            self, 
            cluster_radius : int, 
            min_spot : int
            ) -> Tuple[np.ndarray, np.ndarray, dict, dict] :
        """
        Compute clusters using bigfish detection.detect_clusters and seperate coordinates from features.
        """
        
        clustered_spots, clusters = detection.detect_clusters(
            voxel_size=self.voxel_size,
            spots= self.single_layer.data,
            radius=cluster_radius,
            nb_min_spots= min_spot
        )

        clusters_coordinates = clusters[:,:-2]
        clusters_features = {
            "spot_number" : clusters[:,-2],
            "cluster_id" : clusters[:,-1],
        }

        spots_coordinates = clustered_spots[:,:-1]
        spots_features = {
            "cluster_id" : clustered_spots[:,-1]
        }

        return clusters_coordinates, spots_coordinates, clusters_features, spots_features

    def _update_layers(
            self, 
            clusters_coordinates : np.ndarray, 
            spots_coordinates : np.ndarray, 
            clusters_features : dict, 
            spots_features : dict
            ) -> None  :
        """
        Update Points layers inside napari viewer.
        """
        
        #Modify layers
        self.single_layer.data = spots_coordinates
        self.cluster_layer.data = clusters_coordinates
        self.single_layer.features.loc[:,["cluster_id"]] = spots_features['cluster_id']
        self.cluster_layer.features.loc[:,["cluster_id"]] = clusters_features['cluster_id']
        self.cluster_layer.features.loc[:,["spot_number"]] = clusters_features['spot_number']

        self.cluster_layer.selected_data.clear()
        self.single_layer.refresh()
        self.cluster_layer.refresh()

        

    def _create_widget(self):

        @magicgui(
                call_button= "Relaunch Clustering",
                auto_call= False
        )
        def relaunch_clustering(
            cluster_radius : int = self.cluster_radius,
            min_spot : int = self.min_spot,
        ) :
            clusters_coordinates, spots_coordinates, clusters_features, spots_features = self._compute_clusters(cluster_radius=cluster_radius, min_spot=min_spot)
            self._update_layers(clusters_coordinates, spots_coordinates, clusters_features, spots_features )
            self.cluster_radius = cluster_radius
            self.min_spot = min_spot

        return relaunch_clustering

class ClusterCreator(ClusterWidget) :
    """
    Create a cluster containing all and only selected spots located at the centroid of selected points.
    """
    def __init__(self, cluster_layer, single_layer):
        super().__init__(cluster_layer, single_layer)

    def _create_widget(self):

        @magicgui(
                call_button= "Create Cluster",
                auto_call=False
        )
        def create_foci() -> None :
            selected_spots_idx = pd.Index(list(self.single_layer.selected_data))
            free_spots_idx : pd.Index = self.single_layer.features.loc[self.single_layer.features['cluster_id'] == -1].index
            selected_spots_idx = selected_spots_idx[selected_spots_idx.isin(free_spots_idx)]

            spot_number = len(selected_spots_idx)
            if spot_number == 0 :
                print("To create a cluster please select at least 1 spot")
            else :
                
                #Foci creation
                spots_coordinates = self.single_layer.data[selected_spots_idx]
                new_cluster_id = self.cluster_layer.features['cluster_id'].max() + 1
                centroid = list(spots_coordinates.mean(axis=0).round().astype(int))

                self.cluster_layer.data = np.concatenate([
                    self.cluster_layer.data,
                    [centroid]
                ], axis=0)
                
                last_index = len(self.cluster_layer.data) - 1
                self.cluster_layer.features.loc[last_index, ['cluster_id']] = new_cluster_id
                self.cluster_layer.features.loc[last_index, ['spot_number']] = spot_number

                #Update spots cluster_id
                self.single_layer.features.loc[selected_spots_idx,["cluster_id"]] = new_cluster_id
        
        return create_foci

@register_cluster_wizard
class ClusterInspector :
    """
    Listen to event on cluster layer to color spots belonging to clusters in green
    """
    def __init__(self, single_layer : Points, cluster_layer : Points):
        self.single_layer = single_layer
        self.cluster_layer = cluster_layer
        self.start_listening()

    def reset_single_colors(self) -> None:
        self.single_layer.face_color = [0,0,0,0] #transparent
        self.single_layer.refresh()

    def start_listening(self) :

        def color_single_molecule_in_foci() -> None:
            self.reset_single_colors()
            selected_cluster_indices = self.cluster_layer.selected_data
            for idx in selected_cluster_indices :
                selected_cluster = self.cluster_layer.features.at[idx,"cluster_id"]
                belonging_single_idex = self.single_layer.features.loc[self.single_layer.features['cluster_id'] == selected_cluster].index.to_numpy()
                self.single_layer.face_color[belonging_single_idex] = [0,1,0,1] #Green
                self.single_layer.refresh()

        self.cluster_layer.selected_data.events.items_changed.connect(color_single_molecule_in_foci)

@register_cluster_wizard
class ClusterEraser(ClusterWizard) :
    """
    When a foci is deleted, update spots feature table accordingly.
    """

    def __init__(self, single_layer, cluster_layer):
        super().__init__(single_layer, cluster_layer)

    def start_listening(self):
        self.original_remove_selected = self.cluster_layer.remove_selected
    
        def remove_selected_cluster() :
            selected_cluster = self.cluster_layer.selected_data
            for cluster_idx in selected_cluster : #First we update spots data
                cluster_id = self.cluster_layer.features.at[cluster_idx, "cluster_id"]
                self.single_layer.features.loc[self.single_layer.features['cluster_id'] == cluster_id, ['cluster_id']] = -1
            
            self.original_remove_selected() # Then we launch the usual napari method
        
        self.cluster_layer.remove_selected = remove_selected_cluster

@register_cluster_wizard
class ClusterAdditionDisabler(ClusterWizard) :
    """
    Remove the action when user uses points addition tool for Foci, forcing him to use the FociCreator tool to add new cluster.
    """

    def __init__(self, single_layer, cluster_layer):
        super().__init__(single_layer, cluster_layer)
    
    def start_listening(self):

        def print_excuse(*args, **kwargs):
            print("Spot addition is disabled for cluster layer. Use the foci creation tool below after selecting spots you want to cluster")

        self.cluster_layer.add = print_excuse

@register_cluster_wizard
class SingleEraser(ClusterWizard) :
    """
    When a single is deleted, update clusters feature table accordingly
    """

    def __init__(self, single_layer, cluster_layer):
        super().__init__(single_layer, cluster_layer)

    def start_listening(self):
        self._origin_remove_single = self.single_layer.remove_selected

        def delete_single(*args, **kwargs) :
            selected_single_idx = list(self.single_layer.selected_data)
            modified_cluster_ids = self.single_layer.features.loc[selected_single_idx, ["cluster_id"]].to_numpy().flatten()

            print(np.unique(modified_cluster_ids, return_counts=True))
            for cluster_id, count in zip(*np.unique(modified_cluster_ids, return_counts=True)): # Then update number of spots in cluster
                    if cluster_id == -1 : continue
                    new_spot_number = len(self.single_layer.features.loc[self.single_layer.features['cluster_id'] == cluster_id]) - count #minus number of spot with this cluster id we remove
                    print("new spot number : ", new_spot_number)
                    print('target cluster id : ', cluster_id)
                    self.cluster_layer.features.loc[self.cluster_layer.features['cluster_id'] == cluster_id, ["spot_number"]] = new_spot_number
            self._origin_remove_single()
            self.cluster_layer.events.features()
        
        self.single_layer.remove_selected = delete_single


@register_cluster_wizard
class ClusterCleaner(ClusterWizard) :
    """
    Deletes clusters if they drop to 0 single molecules.
    """

    def __init__(self, single_layer, cluster_layer):
        super().__init__(single_layer, cluster_layer)

    def start_listening(self):

        def delete_empty_cluster() :
            drop_idx = self.cluster_layer.features[self.cluster_layer.features['spot_number'] == 0].index
            
            if len(drop_idx) > 0 :
                print("Removing {} empty cluster(s)".format(len(drop_idx)))
                self.cluster_layer.data = np.delete(self.cluster_layer.data, drop_idx, axis=0)
                self.cluster_layer.refresh()

        self.cluster_layer.events.features.connect(delete_empty_cluster)



#Detection widgets
class BackgroundRemover(NapariWidget) :
    def __init__(
        self, 
        signal : Image, 
        voxel_size : tuple,
        other_image : np.ndarray = None, 
        ) :
        
        self.other_image = other_image
        self.signal_layer = signal
        self.signal_data_raw = np.array(signal.data)
        self.voxel_size = voxel_size
        self.scale = compute_anisotropy_coef(self.voxel_size)

        self.signal_args = {
                "name" : "raw signal",
                "colormap" : 'green',
                "scale" : self.scale,
                "blending" : 'additive'
            }

        self.events = EmitterGroup(source=self.signal_layer, auto_connect=False, background_substraction_event = None)

        super().__init__()
        if self.other_image is None : self.disable_channel() #Image stack is None when image stack is not is_multichannel
        self.reset_widget = self._create_reset_button()

    def disable_channel(self) :
        self.widget.channel.enabled = False

    def _create_widget(self) :
        @magicgui(
            channel = {'min' : 0, 'max' : 0 if self.other_image is None else self.other_image.shape[0] - 1},
            max_trial = {'min' : 0},
        )
        def remove_background(
            background_path : Path,
            channel : int,
            max_trial : int = 100,
        )-> LayerDataTuple :

            print("Substracting background ...", end="", flush=True)
            self.gui = remove_background

            if os.path.isfile(background_path) :
                background = open_image(str(background_path))
            elif self.other_image is None :
                raise FileNotFoundError(f"{background_path} is not a valid file.")
            else :
                background = self.other_image[channel]
            if not background.shape == self.signal_data_raw.shape : raise ValueError(f"Shape missmatch between signal and background : {self.signal_data_raw.shape} ; {background.shape}")

            result, score = remove_autofluorescence_RANSACfit(
                signal=self.signal_data_raw.copy(),
                background=background,
                max_trials=max_trial
            )

            print("\rBackground substraction done.")
            self.events.background_substraction_event(new_signal_array = result)
            
            return (result, self.signal_args, 'image')
        return remove_background
    
    def _create_reset_button(self) :

        @magicgui(call_button= "Reset signal")
        def reset_signal() -> LayerDataTuple :
            self.events.background_substraction_event(new_signal_array = self.signal_data_raw)
            return (self.signal_data_raw, self.signal_args, 'image')
        return reset_signal


class SpotDetector(NapariWidget) :
    """
    Widget aimed at helping user to set detection parameters : threshold, spot radius and so on...
    """

    def __init__(
        self,
        image: np.ndarray,
        default_threshold : int,
        default_spot_size : tuple,
        default_kernel_size : tuple,
        default_min_distance : tuple,
        voxel_size : tuple,
        background_remover_instance : BackgroundRemover,
        ) :
        
        self.image = image
        self.voxel_size = voxel_size
        self.dim = len(voxel_size)
        self.default_threshold = default_threshold
        self.spot_radius = default_spot_size
        self.kernel_size = default_kernel_size
        self.min_distance = default_min_distance
        self._update_filtered_image()
        self.maximum_threshold = self.filtered_image.max()
        self.do_update = False
        
        super().__init__()
        background_remover_instance.events.background_substraction_event.connect(self.on_background_updated)

    def _update_filtered_image(self) :

        print("Re-computing filtered image with new parameters : ...", end="", flush=True)
        self.filtered_image = _apply_log_filter(
            image=self.image,
            voxel_size=self.voxel_size,
            spot_radius=self.spot_radius,
            log_kernel_size=self.kernel_size
        )
        print("\rRe-computing filtered image with new parameters : done")

        self.local_maxima = _local_maxima_mask(
            image_filtered=self.filtered_image,
            voxel_size=self.voxel_size,
            spot_radius=self.spot_radius,
            minimum_distance=self.min_distance
        )

    def _create_widget(self) :
        
        dim = len(self.voxel_size)
        if dim == 2 :
            tuple_hint = Tuple[int,int]
        else :
            tuple_hint = Tuple[int,int,int]

        if not self.default_threshold is None : 
            default_threshold = min(self.default_threshold, self.filtered_image.max())
        else :
            default_threshold = None

        @magicgui(
            threshold = {"widget_type" : SpinBox, "min" : 0, "value" : default_threshold, "max" : self.filtered_image.max() + 1},
            spot_radius = {"label" : "spot radius(zyx)", "value" : self.spot_radius},
            kernel_size = {"label" : "LoG kernel size(zyx)"},
            minimum_distance = {"label" : "Distance min between spots"},
        )
        def find_spots(
            threshold : int,
            spot_radius : tuple_hint,
            kernel_size : tuple_hint,
            minimum_distance : tuple_hint,
        ) -> List[LayerDataTuple] :

            if (np.array(spot_radius) < 0).any() :
                raise ValueError("Spot radius : set value > 0 (0 to ignore argument)")
            
            if (np.array(kernel_size) < 0).any() :
                raise ValueError("Spot radius : set value > 0 (0 to ignore argument)")
            
            if (np.array(minimum_distance) < 0).any() :
                raise ValueError("Spot radius : set value > 0 (0 to ignore argument)")
            
            if isinstance(spot_radius, tuple) :
                if not all(spot_radius) : spot_radius = None #any value set to 0
            if isinstance(kernel_size, tuple) :
                if not all(kernel_size) : kernel_size = None #any value set to 0
            if isinstance(minimum_distance, tuple) :
                if not all(minimum_distance) : minimum_distance = None #any value set to 0

            if spot_radius != self.spot_radius :
                self.spot_radius = spot_radius
                self.do_update = True
            if kernel_size != self.kernel_size :
                self.kernel_size = kernel_size
                self.do_update = True
            if minimum_distance != self.min_distance :
                self.min_distance = minimum_distance
                self.do_update = True
            
            try :
                if self.do_update :
                    self._update_filtered_image()
                    self.do_update = False
                    self.widget.threshold.max = self.filtered_image.max() + 1
                
                print("Computing automated threshold : ...", end="", flush=True)
                if threshold == 0 :
                    threshold = automated_threshold_setting(
                        self.filtered_image,
                        mask_local_max=self.local_maxima
                    )
                    self.widget.threshold.value = threshold
                print("\rComputing automated threshold : done.")

                spots = spots_thresholding(
                image=self.filtered_image,
                mask_local_max=self.local_maxima,
                threshold=threshold
                )[0]
            except ValueError as e :
                print(str(e))


            scale = compute_anisotropy_coef(self.voxel_size)
    
            spot_layer_args = {
                'size': 5, 
                'scale' : scale, 
                'face_color' : 'transparent', 
                'border_color' : 'red', 
                'symbol' : 'disc', 
                'opacity' : 0.7, 
                'blending' : 'translucent', 
                'name': 'single spots',
                'visible' : True,
                }

            filtered_image_layer_args = {
                "colormap" :  'gray',
                "scale" : scale,
                "blending" : 'additive',
                "name" : "filtered image"
            }

            return [
                    (self.filtered_image, filtered_image_layer_args, 'image'),
                    (spots, spot_layer_args, 'points')
                    ]

        return find_spots

    def on_background_updated(self, event):
        print("Background was updated — recomputing filtered image...")
        self.image = event.new_signal_array
        self.do_update = True
        self.widget()

    def get_detection_parameters(self) :
        detection_parameters = {"threshold" : self.widget.threshold.value}
        if self.spot_radius is not None :
            detection_parameters.update({
            "spot_size" : self.spot_radius,
            "spot_size_z" :  self.spot_radius[0] if self.dim == 3 else None,
            "spot_size_y" :  self.spot_radius[0 + (self.dim==3)],
            "spot_size_x" :  self.spot_radius[1 + (self.dim==3)]
            })
        if self.kernel_size is not None :
            detection_parameters.update({
            "log_kernel_size" : self.kernel_size,
            "log_kernel_size_z" : self.kernel_size[0] if self.dim == 3 else None,
            "log_kernel_size_y" : self.kernel_size[0 + (self.dim==3)],
            "log_kernel_size_x" : self.kernel_size[1 + (self.dim==3)]
            })
        if self.min_distance is not None :
            detection_parameters.update({
            "minimum_distance" : self.min_distance,
            "minimum_distance" : self.min_distance[0] if self.dim == 3 else None,
            "minimum_distance" : self.min_distance[0 + (self.dim==3)],
            "minimum_distance" : self.min_distance[1 + (self.dim==3)],
            })

        return detection_parameters
        
        



class DenseRegionDeconvolver(NapariWidget) :
    """
    Widget for interactive detection. Create 2 layes : Labels layer representing dense region that could be deconvoluted and Points layer with deconvoluted spots
    """
    def __init__(
        self,
        image : Image,
        spots : Points, 
        alpha : float, 
        beta : float, 
        gamma : float,
        spot_radius : tuple,
        kernel_size : tuple,
        voxel_size : tuple
        ) :
        
        self.image = image
        self.alpha = alpha
        self.beta = beta
        self.gamma = gamma
        self.spots = spots
        self.spot_radius = spot_radius
        self.kernel_size = kernel_size
        self.voxel_size = voxel_size
        self.dim = len(voxel_size)
        self.update_dense_regions()
        super().__init__()

    def update_dense_regions(self) :
        dense_regions, spot_out_regions,max_size = detection.get_dense_region(
            image=self.image.data,
            spots=self.spots.data,
            voxel_size = self.voxel_size,
            beta=self.beta,
            spot_radius=self.spot_radius
        )
        del spot_out_regions,max_size

        mask = np.zeros(shape=self.image.data.shape, dtype= np.int16)
        for label, region in enumerate(dense_regions) :
            reg_im = region.image
            coordinates = np.argwhere(reg_im)

            if self.dim == 2 :
                y,x = coordinates.T
                min_y,min_x,*_ = region.bbox
                y += min_y
                x += min_x

                mask[y,x] = label + 1

            else :
                z,y,x = coordinates.T
                min_z,min_y,min_x,*_ = region.bbox
                z += min_z
                y += min_y
                x += min_x

                mask[z,y,x] = label + 1

        self.dense_regions = mask

    def _create_widget(self) :

        dim = len(self.voxel_size)
        tuple_hint = Tuple[int,int] if dim == 2 else Tuple[int,int,int]
        tuple_dummy = tuple(0 for i in range(dim))

        @magicgui
        def dense_region_deconvolution(
            alpha : float = self.alpha,
            beta : float = self.beta,
            gamma : float = self.gamma,
            spot_radius : tuple_hint = tuple_dummy if self.spot_radius is None else self.spot_radius,
            kernel_size : tuple_hint = tuple_dummy if self.kernel_size is None else self.kernel_size,
        ) -> List[LayerDataTuple] :

            if (np.array(spot_radius) < 0).any() :
                    raise ValueError("Spot radius : set value > 0 (0 to ignore argument)")
            if (np.array(kernel_size) < 0).any() :
                raise ValueError("kernel size : set value > 0 (0 to ignore argument)")
            
            if isinstance(spot_radius,tuple) :
                if not all(spot_radius) : spot_radius = None #any value set to 0
            if isinstance(kernel_size,tuple) :
                if not all(kernel_size) : kernel_size = None #any value set to 0

            self.do_update = False
            if spot_radius != self.spot_radius :
                self.spot_radius = spot_radius
                self.do_update = True
            if beta != self.beta :
                self.beta = beta
                self.do_update=True
            if self.do_update : 
                print("Updating dense regions...", end="", flush=True)
                self.update_dense_regions()
                print("\rUpdating dense regions : done.")
            self.alpha = alpha
            self.gamma = gamma
            self.kernel_size = kernel_size

            print("Decomposing dense regions...", end="", flush=True)
            spots, _dense_region, _reference_spot = detection.decompose_dense(
                image= self.image.data, 
                spots= self.spots.data, 
                voxel_size=self.voxel_size, 
                spot_radius=self.spot_radius, 
                kernel_size=self.kernel_size, 
                alpha=self.alpha, 
                beta=self.beta, 
                gamma=self.gamma
            )
            print("\rDecomposing dense regions : done")
            del _dense_region, _reference_spot

            scale = compute_anisotropy_coef(self.voxel_size)
            spot_layer_args = {
                'size': 5, 
                'scale' : scale, 
                'face_color' : 'transparent', 
                'border_color' : 'blue', 
                'symbol' : 'disc', 
                'opacity' : 0.7, 
                'blending' : 'translucent', 
                'name': 'decovoluted spots',
                'visible' : True,
                }

            dense_region_args = {
                "scale" : scale,
                "name": "Dense regions",
                "colormap" : ["red"] * self.dense_regions.max()
            }

            return [(self.dense_regions, dense_region_args, 'labels'), (spots, spot_layer_args, 'points')]
        return dense_region_deconvolution

    def get_detection_parameters(self) :
        return {
            "alpha" : self.alpha,
            "beta" : self.beta,
            "gamma" : self.gamma
        }

